"""
Created on 13.03.2015

@author: freu_se

Implementing Morris' algorithm(1991) described in Forrester: "Engineering Design via Surrogate Modelling" chapter 1.3.1.

As example please refer to test/test_service/test_screening.py
"""

import numpy as np


def screeningStdMeans(screeningPlanArray, targetSystemValues, xi, l):
    """Generates a variable elementary effect screening plot

    :param screeningPlan: screening plan with values of [0,1] of the shape (k+1*r, k) (built e.g. with screeningPlan()).
        This array corresponds to sampleX.
    :param targetSystemValues: list or np.ndarray with values of target system at each screeningPlan[i]-point
        This array corresponds to sampleY.
    :param xi: elementary effect step length factor
    :param l: number of discrete levels along each dimension

    :return: tuple with 2 entries: List of means, list of standard deviations
    """
    screeningPlanArray = screeningPlanArray.T
    k = screeningPlanArray.shape[1]
    r = screeningPlanArray.shape[0] // (k + 1)
    t = np.array(targetSystemValues)

    d = np.zeros((k, r))
    screeningPlanChunks = np.vsplit(screeningPlanArray, r)
    for chunkIndex, spChunk in enumerate(screeningPlanChunks):
        for index in range(len(spChunk) - 1):
            parameterIndex = np.nonzero(spChunk[index + 1] - spChunk[index])[0][0]
            if (spChunk[index + 1] - spChunk[index])[parameterIndex] < 0:
                parameterDirection = -1
            else:
                parameterDirection = 1
            d[parameterIndex, chunkIndex] = (
                (t[(index + 1) + chunkIndex * (k + 1)] - t[index + chunkIndex * (k + 1)])
                / (xi / (l - 1.0))
                * parameterDirection
            )

    standardDeviations = np.std(d, 1)
    means = np.mean(d, 1)

    meansStar = np.mean(np.abs(d), 1)
    """
    extension from [1]F. Campolongo, J. Cariboni, und A. Saltelli, "An effective screening design for sensitivity analysis of large models", Environmental modelling & software, Bd. 22, Nr. 10, S. 1509-1518, 2007.
    also on page 111 in [1]A. Saltelli u. a., Global Sensitivity Analysis: The Primer, 1. Aufl. Chichester, England; Hoboken, NJ: John Wiley & Sons, 2008.
    """

    return (means, standardDeviations, meansStar)


def plotMeansAndStd(means, standardDeviations, parameterNames):
    """This method creates a plot of the means and standard deviations of a screening"""
    try:
        # these are imports from delis just for output reasons. If you do not have delis installed or in your
        # pythonpath, this method is skipped. You can simply create your own output plot as seen in the book
        from delis.model.geometry.translate import Translation
        from delis.service.geometrywriter import MatplotlibWriter
    except:
        return
    mplWriter = MatplotlibWriter("", "Elementary effects distribution plot", axes="auto")
    mplWriter.scalePlot = False
    mplWriter.printID = True
    mplWriter.textOffset = min([max(means), max(standardDeviations)]) * 1e-2
    mplWriter.marker = "+"
    mplWriter.ax.set_xlabel("Sample Means")
    mplWriter.ax.set_ylabel("Sample Standard Deviations")

    for mean, std, name in zip(means, standardDeviations, parameterNames):
        mplWriter.point(Translation([mean, std, 0], id=name))

    mplWriter.end()



# Copyright (C) 2013 Deutsches Zentrum fuer Luft- und Raumfahrt(DLR, German Aerospace Center) <www.dlr.de>
"""
Created on 07.05.2014

@author: freu_se
"""


class DelisMMError(Exception):
    """classdocs"""



"""
Created on 24.01.2017

@author: freu_se
"""

import copy
import getpass
import logging
import os
import pickle
import time
import traceback
from configparser import ConfigParser
from multiprocessing import Pool, Process, Queue, Value, cpu_count, current_process

import numpy as np
import pandas as pd
from patme.service.decorators import memoize
from patme.service.logger import log, resetLoggerToNewRunDir
from patme.service.systemutils import getTimeString
from patme.sshtools.cara import numberOfClusterJobsAvailable

from delismm.service.exception import DelisMMError


@memoize
def getY(sampleX, targetFunctionInstance, verbose=True, runDir=None, resumeSamples=False):
    """Method that returns function values at the sample points.

    It is the main entry point to calculate samples. There are several variants for calculation.

    1. calculate serial
    #. calculate parallel. For this, targetFunctionInstance must have the attribute doParallelization=[clusterqueue, local] or one of them.
    #. calculate remote. Performed if, targetFunctionInstance.doRemoteCall==True must have the attribute doRemoteCall=True

    :param sampleX: p,n array with samples not normalized
    :param targetFunctionInstance: instance of a delismm.model.customsystemfunction class
        or a sub class of it.
    :param verbose: Flag if the calculation should be output to the commandline
    :param runDir: directory of the run. If present, for serial runs, the results are written after each sample
    :param resumeSamples: Flag if sampleY in runDir should be resumed. In this case, sampleY.pickle is read and the rest of the calculation resumed.
        Only works for serial or local parallel runs with given runDir.
        If None, a new run is started
    :return: function values at the sample points
    """
    if sampleX is None:
        log.warning("SampleX not given in getYStatic. Returning")
        return None
    sampleX = np.array(sampleX)

    # parallelization state
    doRemoteCall = hasattr(targetFunctionInstance, "doRemoteCall") and targetFunctionInstance.doRemoteCall
    clusterQueueParallelization, localParalellization, asyncParallelization = False, False, False
    if hasattr(targetFunctionInstance, "doParallelization"):
        clusterQueueParallelization = "clusterqueue" in targetFunctionInstance.doParallelization
        localParalellization = "local" in targetFunctionInstance.doParallelization
        asyncParallelization = "async" in targetFunctionInstance.doParallelization

    # remote runs
    try:
        if clusterQueueParallelization and doRemoteCall:
            return getYParallel(sampleX, targetFunctionInstance, verbose)
        if doRemoteCall:
            return remoteCaller(sampleX, targetFunctionInstance, verbose)

    # local runs
    except Exception as e:
        log.error(
            "Received an error while performing the function call remotely. Trying local run instead. Original error:\n"
            + str(e)
        )

        return _getYLocalTemporary(sampleX, targetFunctionInstance, verbose)
    if localParalellization:
        return getYParallel(sampleX, targetFunctionInstance, verbose, resumeSamples=resumeSamples)
    elif asyncParallelization:
        return getYParallelAsync(sampleX, targetFunctionInstance, verbose)
    return getYSerial(sampleX, targetFunctionInstance, verbose, runDir, resumeSamples=resumeSamples)


def getYSerial(sampleX, targetFunctionInstance, verbose=True, runDir=None, resumeSamples=False):
    """Runs the target function in serial

    For a parameter documentation, refer to getY()"""
    from delismm.model.doe import AbstractDOE, DOEfromFile

    if resumeSamples:
        sampleYFile = os.path.join(runDir, "sampleY.pickle")
        sampleY = AbstractDOE.ysFromFile(sampleYFile, targetFunctionInstance)
        sampleX = sampleX[:, len(sampleY) :]
        log.info("Resuming samples. Already calculated: " + str(len(sampleY)))
    else:
        sampleY = []
    numberOfSamples = sampleX.shape[1]
    if verbose:
        log.info("Calculating samples.")
    if runDir:
        sampleXFilename = os.path.join(runDir, "sampleX_bounds.txt")
        if not os.path.exists(sampleXFilename):
            AbstractDOE.xToFileStatic(sampleXFilename, sampleX)
    for sampleNumber in range(numberOfSamples):
        sampleY.append(targetFunctionInstance(sampleX[:, sampleNumber]))
        if verbose and divmod(sampleNumber + 1, 10)[1] == 0:
            log.info("Calculated " + str(sampleNumber + 1) + " samples.")
        if runDir:
            AbstractDOE.yToFile(os.path.join(runDir, "sampleY.txt"), targetFunctionInstance, sampleY)
    return sampleY


def _getYLocalTemporary(sampleX, targetFunctionInstance, verbose):
    """calculates the samples locally if a remote call failed. The parameters are adpated only temporary"""
    targetFunctionInstance.doRemoteCall = False
    clusterqueue = (
        True
        if hasattr(targetFunctionInstance, "doParallelization")
        and "clusterqueue" in targetFunctionInstance.doParallelization
        else False
    )
    if clusterqueue:
        targetFunctionInstance.doParallelization.remove("clusterqueue")

    sampleY = getY(sampleX, targetFunctionInstance, verbose)

    if clusterqueue:
        targetFunctionInstance.doParallelization.append("clusterqueue")
    targetFunctionInstance.doRemoteCall = True

    return sampleY


def getYParallel(sampleX, targetFunctionInstance, verbose=True, resumeSamples=False):
    """Runs the target function in parallel.

    It determines the number of processes either by the number of cores or by the
    number of cluster jobs. Then for each parallel process, a copy of the targetfunction
    running in a subdirectory is created. Lastly, the samples are split for each process.
    After the parallel run, the results are put together and returned.
    """
    numberOfProcesses = _getNumberOfProcesses(sampleX.shape[1], targetFunctionInstance)
    targetFunctionCopies = _getTargetFunctionCopies(targetFunctionInstance, numberOfProcesses, resumeSamples)
    processInputs = _getParallelRunInputSamples(
        sampleX, targetFunctionCopies, numberOfProcesses, verbose, resumeSamples
    )

    # run parallel
    log.info("Start parallel run with " + str(numberOfProcesses) + " processes")
    with Pool(processes=numberOfProcesses) as pool:
        sampleYmultiprocesses = pool.map(calledParallelExecution, processInputs)

    # put results together
    sampleY = []
    for sampleYmultiprocess in sampleYmultiprocesses:
        sampleY.extend(sampleYmultiprocess)
    log.info("parallel run done")

    return sampleY


manualMinParallelProcesses = None


def _getNumberOfProcesses(sampleLength, targetFunctionInstance):
    """Calculates the number of processes that might be run. Processes for the cluster
    are dependent on the number of available cluster jobs. Local parallelization depends
    on the number of cores"""
    if "clusterqueue" in targetFunctionInstance.doParallelization:
        numberOfProcesses = numberOfClusterJobsAvailable() - 1
        numberOfProcesses = np.max([numberOfProcesses, 2])
        numberOfProcesses = np.min(
            [numberOfProcesses, 8]
        )  # use a maximum of 8 cluster processes. Otherwise protocol banner error might occur
    else:
        numberOfProcesses = int(np.ceil(cpu_count() * 2 / 3))
    return int(
        np.min([sampleLength, numberOfProcesses, manualMinParallelProcesses if manualMinParallelProcesses else np.inf])
    )


def _getTargetFunctionCopies(targetFunctionInstance, numberOfProcesses, resumeSamples=False):
    """creates copies of the given target function. It removes the parallelization mode, that is
    used (clusterqueue or local). For each new function a subdirectory is created for a run
    independently to the other target functions."""
    targetFunctionCopies = []
    parallelDirName = "parallel_"
    runDir = targetFunctionInstance.runDir if hasattr(targetFunctionInstance, "runDir") else None
    foldersFound = []
    if runDir is not None and resumeSamples:
        # look for folders like with parallelDirName
        for dirName in os.listdir(runDir):
            dirPath = os.path.join(runDir, dirName)
            if os.path.isdir(dirPath) and parallelDirName in dirName:
                foldersFound.append(dirPath)
        if not len(foldersFound) == numberOfProcesses:
            raise DelisMMError(
                "Number of parallel folders found in runDir does not match number of processes. Found: "
                + f"{len(foldersFound)}, Expected: {numberOfProcesses}"
            )
    for processNumber in range(numberOfProcesses):
        targetFunctionCopy = copy.deepcopy(targetFunctionInstance)
        if "clusterqueue" in targetFunctionCopy.doParallelization:
            targetFunctionCopy.doParallelization.remove("clusterqueue")
        elif "local" in targetFunctionCopy.doParallelization:
            targetFunctionCopy.doParallelization.remove("local")
        else:
            targetFunctionCopy.doParallelization.remove("async")
        # reset rundir
        if resumeSamples:
            targetFunctionCopy.runDir = foldersFound[processNumber]
        elif targetFunctionCopy.runDir is not None:
            targetFunctionCopy.runDir = (
                os.path.abspath(os.path.join(runDir, parallelDirName + getTimeString())) + "_" + str(processNumber)
            )
            if not os.path.exists(targetFunctionCopy.runDir):
                os.makedirs(targetFunctionCopy.runDir)
        targetFunctionCopies.append(targetFunctionCopy)
    return targetFunctionCopies


def _getParallelRunInputSamples(sampleX, targetFunctionCopies, numberOfProcesses, verbose, resumeSamples=False):
    """Create a list of 3-tuples. Each list entry has the input for each parallel process to be run.
    A tuple consists of (sub-sampleX, targetFunctionCopy, verbose)

    Each tuple will be used for the method "calledParallelExecution"
    """
    # split samples
    splitSampleLength = int(np.ceil(sampleX.shape[1] / numberOfProcesses))
    splitSamples = [
        sampleX[:, splitSampleLength * count : splitSampleLength * (count + 1)]
        for count in range(numberOfProcesses)
        if splitSampleLength * count < sampleX.shape[1]
    ]

    # setup input list
    return list(
        zip(
            splitSamples,
            targetFunctionCopies,
            range(numberOfProcesses),
            [verbose] * numberOfProcesses,
            [resumeSamples] * numberOfProcesses,
        )
    )


def getYParallelAsync(sampleX, targetFunctionInstance, verbose=True):
    """This function performs a parallel execution of the target function.
    The targetfunction must have these attributes set:

    - asyncWaitTime
    - asyncMaxProcesses
    - getNumberOfNewJobs()

    Please refer to customsystemfunction for a description of those.

    The run is performed asynchronously, so that it respects if a new job can
    be run (by a call to getNumberOfNewJobs()) e.g. due to license restrictions.
    A task creation process is generated which submits a job if getNumberOfNewJobs() > 0.
    These jobs of single samples are taken by the worker processes, that perform the
    calculation and return the result and sample index(since it must be put in the correct
    order afterwards).
    """
    if targetFunctionInstance.asyncMaxProcesses is None:
        raise DelisMMError("Maximum number of processes must be set.")
    numberOfSamples = sampleX.shape[1]
    numberOfWorkerProcesses = np.min([targetFunctionInstance.asyncMaxProcesses, numberOfSamples])

    # Create queues
    taskQueue = Queue()
    resultQueue = Queue()
    numberOfRunningJobs = Value("I", lock=True)

    # Submit tasks in a task process that only puts new jobs in the taskQueue, if it is possible

    runDir = targetFunctionInstance.runDir if hasattr(targetFunctionInstance, "runDir") else None
    processInputs = _getParallelRunInputSamples(
        sampleX, [targetFunctionInstance] * numberOfSamples, numberOfSamples, verbose
    )
    process = Process(
        target=taskCreatorAsync,
        args=(taskQueue, resultQueue, processInputs, runDir, numberOfWorkerProcesses, numberOfRunningJobs),
    )
    process.start()
    process.join()

    # Start worker processes
    targetFunctionCopies = _getTargetFunctionCopies(targetFunctionInstance, numberOfWorkerProcesses)
    for targetFunctionCopy in targetFunctionCopies:
        process = Process(target=workerAsync, args=(taskQueue, resultQueue, targetFunctionCopy, numberOfRunningJobs))
        process.start()
        process.join()

    resultArray = []
    # wait for results and get results
    for _ in range(numberOfSamples):
        resultArray.append(resultQueue.get())
    # sort results
    resultDf = pd.DataFrame(resultArray, columns=["results", "indexes"])
    resultDf.sort_values(by="indexes", inplace=True)

    log.info("parallel run done")

    return resultDf["results"].tolist()


def taskCreatorAsync(taskQueue, resultQueue, processInputs, runDirBase, numberOfWorkerProcesses, numberOfRunningJobs):
    """Function run by the asynchronous task generator processes

    :param runDir: directory where the base process is working - a new folder for own logging files is created
    """

    def stopWorkerProcesses(onError=False):
        log.debug("stop worker processes")
        for _ in range(numberOfWorkerProcesses):
            taskQueue.put("STOP")
        if onError:
            for sampleIndex in range(len(processInputs)):
                # puts None in the result Queue, indicating an error
                resultQueue.put([None, sampleIndex])

    def getMaxNewJobs(runDir, verboseOutput=False):
        logLevel = log.INFO if verboseOutput else log.DEBUG
        maxjobs, minjobs, unusedLicense = readAsyncTaskGenConfig(runDir)
        newJobsByLicense = targetFunctionInstance.getNumberOfNewJobs()
        maxNewJobs = _evaluateMaxNewJobArgs(
            maxjobs, minjobs, unusedLicense, numberOfRunningJobs.value, newJobsByLicense
        )
        log.log(
            logLevel,
            "MaxNewJobs: {}. All job determining parameters are: {}".format(
                maxNewJobs,
                list(
                    zip(
                        ["newJobsByLicense", "maxjobs", "minjobs", "unusedLicense", "numberOfRunningJobs", "waitTime"],
                        [
                            newJobsByLicense,
                            maxjobs,
                            minjobs,
                            unusedLicense,
                            numberOfRunningJobs.value,
                            targetFunctionInstance.asyncWaitTime,
                        ],
                    )
                ),
            ),
        )
        return maxNewJobs

    try:
        runDir = None
        if runDirBase is not None:
            runDir = os.path.abspath(os.path.join(runDirBase, "parallel_{}_taskGen".format(getTimeString())))
            _resetLogger(runDir)
            writeAsyncTaskGenConfig(runDir)

        for sampleX, targetFunctionInstance, sampleIndex, verbose, _ in processInputs:
            newJobCount = getMaxNewJobs(runDir, verboseOutput=False)
            log.info(
                "Try send and start sample #{}. Actual max number of new jobs: {}".format(sampleIndex, newJobCount)
            )
            iterationIndex = 0
            while newJobCount == 0:
                verbose = iterationIndex % 12 == 0
                if verbose:
                    log.info(
                        "Waiting for next possible calculation. Actual max number of new jobs: {}".format(newJobCount)
                    )
                time.sleep(5)
                newJobCount = getMaxNewJobs(runDir, verboseOutput=verbose)
                iterationIndex += 1
            log.debug("put sample in queue. sample: \n" + str(sampleX))

            #             if sampleIndex != 257:continue

            taskQueue.put([sampleX, sampleIndex, verbose])
            # wait for a certain time specified by the target function (e.g. to start the previous process)
            time.sleep(targetFunctionInstance.asyncWaitTime)
        log.info("All tasks submitted")
    except Exception as e:
        stopWorkerProcesses(True)
        log.error("Got an exception: " + str(e))
        raise

    stopWorkerProcesses()


def workerAsync(taskQueue, resultQueue, targetFunctionCopy, numberOfRunningJobs):
    """Function run by the asynchronous worker processes"""
    firstIteration = True
    for sampleX, sampleIndex, verbose in iter(taskQueue.get, "STOP"):
        with numberOfRunningJobs.get_lock():
            numberOfRunningJobs.value += 1
        if firstIteration:
            _resetLogger(targetFunctionCopy.runDir)
            firstIteration = False
        log.info(f"calculate sample #{sampleIndex} in {current_process().name} and these samples: {sampleX}")
        try:
            result = _callGetYInNewProcess(sampleX, targetFunctionCopy, sampleIndex, verbose)
            resultQueue.put([result[0], sampleIndex])
        except:
            resultQueue.put([None, sampleIndex])
        with numberOfRunningJobs.get_lock():
            numberOfRunningJobs.value -= 1
    log.info('Subprocess done with "{}" and processId {}'.format(str(current_process().name), str(os.getpid())))


def calledParallelExecution(args):
    """method that is called from the synchronous parallel process. Resets the logger and calls getY"""
    log.info("in calledParallelExecution")
    sampleX, targetFunctionInstance, sampleSetIndex, verbose, resumeSamples = args
    _resetLogger(targetFunctionInstance.runDir)
    sampleY = _callGetYInNewProcess(sampleX, targetFunctionInstance, sampleSetIndex, verbose, resumeSamples)
    log.info('Subprocess done with "{}" and processId {}'.format(str(current_process().name), str(os.getpid())))
    return sampleY


def _callGetYInNewProcess(sampleX, targetFunctionInstance, sampleSetIndex, verbose, resumeSamples=False):
    """doc"""
    log.info('Call getY in "%s" with processId %s' % (str(current_process().name), str(os.getpid())))
    log.debug("calculate sample set #{} and these samples: {}".format(sampleSetIndex, str(sampleX)))
    try:
        sampleY = getY(
            sampleX, targetFunctionInstance, verbose, runDir=targetFunctionInstance.runDir, resumeSamples=resumeSamples
        )
    except Exception as e:
        log.error(
            "\n".join(
                [
                    "Exception raised in " + str(current_process().name) + " with this message:",
                    str(e),
                    traceback.format_exc(),
                ]
            )
        )
        logging.shutdown()
        raise
    return sampleY


def _resetLogger(runDir):
    """resets the logger when a new process is started to prevent the processes to log in to the same file"""
    try:
        os.makedirs(runDir, exist_ok=True)
        resetLoggerToNewRunDir(runDir)
    except:
        # if delis is not on the search path, the logger will not be reset, since there is no file handler in the logger
        pass


taskGeneratorConfigName = "taskGenerator.config"
configSection = "AsyncTaskGeneratorOptions"


def writeAsyncTaskGenConfig(runDir, maxjobs=1000000, minjobs=1, unusedLicenses=1):
    """writes a config file defining configuration options for the asynchronous task generator"""
    configString = _getAsyncTaskGenConfig(maxjobs, minjobs, unusedLicenses)

    with open(os.path.join(runDir, taskGeneratorConfigName), "w") as f:
        f.write(configString)


def _getAsyncTaskGenConfig(maxjobs=1000000, minjobs=1, unusedLicenses=1):
    """writes a config file defining configuration options for the asynchronous task generator"""

    separator = "###################################################"
    configString = """{}
# configuration options for the asynchronous task generator
# hierarchy in case of conflicting values: 1. minJobs, 2.unusedLicenses, 3. maxJobs
{}
[{}]

# maximum number of jobs that may run in parallel
maxJobs = {}

# minimum number of jobs that may run in parallel. The jobs will be sent, without regard of actual license availability
minJobs = {}

# number of licenses that will be kept available to others and will be by this task
unusedLicenses = {}
""".format(
        separator, separator, configSection, maxjobs, minjobs, unusedLicenses
    )
    return configString


def readAsyncTaskGenConfig(runDir):
    """reads the config file from runDir

    :return: tuple: (maxjobs, minjobs, unusedLicense)
    """
    if runDir is None or not os.path.exists(os.path.join(runDir, taskGeneratorConfigName)):
        return 1000000, 0, 0
    with open(os.path.join(runDir, taskGeneratorConfigName)) as f:
        configString = f.read()
    return _readAsyncTaskGenConfig(configString)


def _readAsyncTaskGenConfig(configString):
    """reads the given config string.

    :param configString: string e.g. as generated by _getAsyncTaskGenConfig()

    :return: tuple: (maxjobs, minjobs, unusedLicense)
    """
    parser = ConfigParser(allow_no_value=False)
    parser.read_string(configString)
    configItems = ["maxJobs", "minJobs", "unusedLicenses"]
    configResults = [int(parser[configSection][configItem]) for configItem in configItems]
    return configResults


def _evaluateMaxNewJobArgs(maxjobs, minjobs, unusedLicense, numberOfRunningJobs, newJobsByLicense):
    """doc"""
    maxNewJobs = min(maxjobs - numberOfRunningJobs, newJobsByLicense - unusedLicense)
    maxNewJobs = max(minjobs - numberOfRunningJobs, maxNewJobs, 0)
    return maxNewJobs


def readMaxJobsFile(runDir, verbose=False):
    """reads the file 'maxjobs.txt' if it exists. The file should contain an integer
    defining the maximum number of actual licenses to be used.
    The runDir is usually the task generator process dir: '*_taskGen'"""
    if runDir is None or not os.path.exists(os.path.join(runDir, "maxjobs.txt")):
        return 100000000
    with open(os.path.join(runDir, "maxjobs.txt")) as f:
        line = f.readline()
        try:
            maxJobs = int(line)
            if verbose:
                log.info("Found a max jobs file. Read maxjobs: {}".format(maxJobs))
            return maxJobs
        except:
            log.warning("Found a max jobs file, but could not read the content")
            return 100000000


_getYRemoteInputFileName = "getYInput.pickle"
_getYRemoteResultFileName = "getYResult.pickle"


def remoteCaller(sampleX, targetFunctionInstance, verbose=True):
    """This method handles calling the targetfunction remotely on the institute cluster"""

    # serializing sampleX and targetFunction and put them to a file in targetFucntionInstance.runDir
    tfAndSamplesFile = os.path.join(targetFunctionInstance.runDir, _getYRemoteInputFileName)
    sampleYFile = os.path.join(targetFunctionInstance.runDir, _getYRemoteResultFileName)
    if os.path.exists(tfAndSamplesFile):
        os.remove(tfAndSamplesFile)
    if os.path.exists(sampleYFile):
        os.remove(sampleYFile)

    runDirBak = targetFunctionInstance.runDir  # remotely, the current dir will be targetFunctionInstance.runDir
    targetFunctionInstance.runDir = "."
    with open(tfAndSamplesFile, "wb") as f:
        pickle.dump([sampleX, targetFunctionInstance, verbose], f)
    targetFunctionInstance.runDir = runDirBak

    # call remote
    _callRemote(targetFunctionInstance)

    # read sampleY
    if not os.path.exists(sampleYFile):
        raise FileNotFoundError(
            "The expected sample result does not exist. Possibly, the cluster call failed. Please have a look at the cluster log in this directory: "
            + targetFunctionInstance.runDir
        )
    with open(sampleYFile, "rb") as f:
        sampleY = pickle.load(f)

    if len(sampleY) != sampleX.shape[1]:
        raise DelisMMError("lengths of sampleX and sampleY do not match!")
    return sampleY


def _callRemote(targetFunctionInstance):
    """sends the remote call to the cluster"""
    from patme.sshtools.clustercaller import PythonModuleCaller

    jobName = targetFunctionInstance.name.replace(" ", "_")
    jobCommands = []
    jobCommands += ["/home/%s/delismm/src/delismm/model/samplecalculator.py" % getpass.getuser()]
    caller = PythonModuleCaller(targetFunctionInstance.runDir, jobCommands)
    caller.runRemote(jobName=jobName)


def remoteReader():
    """method called remote (on the cluster) to read the samples and target function, calculate sampleY and write sampleY."""
    try:
        log.info("On remote machine: calculate samples remote")

        log.debug("Read target function and samples")
        runDir = "."  # on the cluster, the process is started in the directory with the input samples
        tfAndSamplesFile = os.path.join(runDir, _getYRemoteInputFileName)
        sampleYFile = os.path.join(runDir, _getYRemoteResultFileName)

        with open(tfAndSamplesFile, "rb") as f:
            sampleX, targetFunctionInstance, verbose = pickle.load(f)

        log.debug("Call getY")
        targetFunctionInstance.doRemoteCall = False
        sampleY = getY(sampleX, targetFunctionInstance, verbose)

        log.debug("Write sampleY")
        with open(sampleYFile, "wb") as f:
            pickle.dump(sampleY, f)

    except Exception as e:
        log.error("\n".join(["Got this error on the remote machine:", str(e), traceback.format_exc()]))
        raise



"""
Created on 13.04.2018

@author: geie_ma
"""

import numpy as np
import scipy as sp
from patme.service.logger import log
from scipy.optimize import differential_evolution

from delismm.model.customsystemfunction import BoundsHandler
from delismm.model.metrics import normalizedRootMeanSquareError
from delismm.model.surrogate import AbstractSurrogate
from delismm.service.exception import DelisMMError


class Rbf(AbstractSurrogate):
    """RBF class - takes at least two arguments: sampleX, sampleY.

    :param sampleX: is an array in which contains the sample points column by column (one parameter configuration = one column)
        Attention: sampleX is scaled to bounds. This module normalizes sampleX itself
    :param sampleY: is an array in which contains the scalar function values at the sample points
    :param lowerBounds: array with lower bounds of length self.p. Defaults to [0, 0, ...]
    :param upperBounds: array with upper bounds of length self.p. Defaults to [1, 1, ...]
    :param regress: type of regression model used (None, 'const', 'linear)
    :param rbfFunction: type of rbf form function ('multiquadric','inverse','gaussian','linear','cubic','quintic','thin_plate')
    :param parameterNames: list with parameter names with length of self.p
    """

    def __init__(
        self,
        sampleX,
        sampleY,
        lowerBounds=None,
        upperBounds=None,
        regress="linear",
        rbfFunction="multiquadric",
        parameterNames=None,
        resultNames=None,
    ):
        """RBF class that combines scipy.interpolate.Rbf with a regression model as well as an automated
        optimization of the regularization parameter as well as epsilon (only needed for multiquadric rbf)"""
        AbstractSurrogate.__init__(self, sampleX, sampleY, lowerBounds, upperBounds, parameterNames, resultNames)

        self.n = self.sampleXNormalized.shape[1]  # n - Number of sample points
        """n is the number of sample points"""
        self.p = self.sampleXNormalized.shape[0]  # p - Number of design variables
        """p is the dimension of the design space/ the number of model parameters"""
        self.regularizationParameter = None
        """parameter that allows relaxed fitting of rbf function (see smooth in scipy.interpolate.Rbf)"""
        self.doRegularizationParameterOpt = True
        """switch that performs the optimization of the regularization parameter"""
        self.numberOfOpt = 1
        """number of optimization runs to determine the regularization parameter"""
        self._scipyRbf = None
        """holds instance of scipy.interpolate.Rbf()"""
        self.rbfFunction = rbfFunction  #'multiquadric','inverse','gaussian','linear','cubic','quintic','thin_plate'
        """rbf form function type"""
        self.epsilon = None
        """parameter epsilon for multiquadric rbf function; epsilon will be estimated if None is provided
        (see scipy.interpolate.Rbf documentation)"""
        self.doEpsilonOpt = True
        """parameter epsilon for multiquadric rbf function can be optimzed instead of using the default value
        (see scipy.interpolate.Rbf documentation)"""
        self._regressCoeff = []
        """list of regression function coefficients"""
        self._sampleYRegressDelta = []
        """sampleY - regressionFunction(sampleX); empty if if self.regress=None"""
        self.name = "rbf"
        """surrogate model name"""

        # =======================================================================
        # regression method
        # can be one of [None, linear, const]
        # =======================================================================
        if regress == "linear":
            self.regress = "linear"
            log.debug("Use linear regression")
        elif regress == "const":
            self.regress = "const"
            log.debug("Use constant regression")
        elif regress == None:
            self.regress = None
            log.debug("Use no regression")
        else:
            raise DelisMMError("Only these values are applicable for parameter regress: [linear, const, None].")

        if self.rbfFunction not in ["multiquadric", "inverse", "gaussian", "linear", "cubic", "quintic", "thin_plate"]:
            raise DelisMMError(
                "Only these values are applicable for parameter regress: "
                "[multiquadric, inverse, gaussian, linear, cubic, quintic, thin_plate]."
            )

    def createSurrogateModel(self):
        """Creates the surrogate model"""
        # create regression model
        if self.regress:
            if self.regress == "linear":
                xData_linReg = self.sampleXNormalized.T
                xData_linReg = np.c_[xData_linReg, np.ones(xData_linReg.shape[0])]
                self._regressCoeff = np.linalg.lstsq(xData_linReg, self.sampleY)[0]
                self._sampleYRegressDelta = self.sampleY - [
                    sum(xData_linReg[pos] * self._regressCoeff) for pos in range(len(xData_linReg))
                ]
            elif self.regress == "const":
                self._regressCoeff = np.array([np.mean(self.sampleY)])
                self._sampleYRegressDelta = self.sampleY - self._regressCoeff[-1]

        # optimize the regularization parameter and epsilon
        if self.doRegularizationParameterOpt:
            optResults = []
            while len(optResults) + 1 <= self.numberOfOpt:
                log.info("Performing %d. parameter optimization for the RBF surrogate model" % (len(optResults) + 1))
                if self.rbfFunction == "multiquadric" and self.doEpsilonOpt:
                    bounds = [(0, 1), (1e-10, 10)]
                else:
                    bounds = [(0, 1)]
                optResult = differential_evolution(self._regularizationParameterFittnessFct, bounds)
                optResults.insert(-1, [optResult.fun] + list(optResult.x))
                if self.numberOfOpt > 1:
                    log.info("CV NRMSE / regParam / epsilon (opt): " + " / ".join([str(pos) for pos in optResults[-1]]))
            minNRMSEindex = list(np.array(optResults)[:, 0]).index(min(np.array(optResults)[:, 0]))
            self.regularizationParameter = optResults[minNRMSEindex][1]
            if self.rbfFunction == "multiquadric" and self.doEpsilonOpt:
                self.epsilon = optResults[minNRMSEindex][2]
            log.info(
                "--> BEST PARAMETERS (CV NRMSE / regParam / epsilon (opt)): "
                + " / ".join([str(pos) for pos in optResults[-1]])
            )
        elif self.regularizationParameter == None:
            raise DelisMMError(
                "If no regularization parameter optimization shall be performed, a regularization"
                "parameter has to be set manually."
            )

        if self.regress:
            yData = self._sampleYRegressDelta
        else:
            yData = self.sampleY
        self._scipyRbf = sp.interpolate.Rbf(
            *self.sampleXNormalized,
            yData,
            function=self.rbfFunction,
            smooth=self.regularizationParameter,
            epsilon=self.epsilon,
        )

    def _regularizationParameterFittnessFct(self, optParams):
        """Fittness function for the regularization parameter optimization.  returning the normalized root mean square error.
        :param optParams: list with optimization parameters [regularizationParameter, epsilon (optional)]
        :return: normalized root mean square error
        """
        if len(optParams) == 2:
            regularizationParameter, eps = optParams[0], optParams[1]
        #             if eps < 1E30 and eps > 0:
        #                 eps = 1E30
        #             if eps > -1E30 and eps < 0:
        #                 eps = -1E30
        else:
            regularizationParameter = optParams[-1]
            eps = None
        ys_rbf = []
        for pos in range(len(self.sampleY)):
            x_temp = np.delete(self.sampleXNormalized, pos, 1)
            if self.regress:
                y_temp = np.delete(self._sampleYRegressDelta, pos, 0)
            else:
                y_temp = np.delete(self.sampleY, pos, 0)
            rbf = sp.interpolate.Rbf(
                *x_temp, y_temp, function=self.rbfFunction, smooth=regularizationParameter, epsilon=eps
            )
            y_rbf = rbf(*self.sampleXNormalized[:, pos])
            ys_rbf.append(y_rbf)
        if self.regress:
            return normalizedRootMeanSquareError(self._sampleYRegressDelta, ys_rbf)
        else:
            return normalizedRootMeanSquareError(self.sampleY, ys_rbf)

    def __call__(self, xData):
        """Returns the function value at pointOfInterest.
        :param pointOfInterest: vector of length self.p within real bounds"""
        xData_Normalized = BoundsHandler.scaleTo01Static(np.array(xData), self.lowerBounds, self.upperBounds)
        return self.callNormalized(xData_Normalized)

    def callNormalized(self, xData_Normalized):
        """same as __call__ but uses normalized input samples"""
        ys = self._scipyRbf(*xData_Normalized)
        if self.regress == "linear":
            ys = ys + sum(np.append(xData_Normalized, 1) * self._regressCoeff)
        if self.regress == "const":
            ys = ys + self._regressCoeff[-1]
        return np.array([float(ys)])



import os
import subprocess

if __name__ == "__main__":
    port = 5007
    subprocess.call(
        [
            "bokeh",
            "serve",
            "--show",
            f"--port={port}",
            f"--allow-websocket-origin=localhost:{port}",
            f"--allow-websocket-origin=fa-121050:{port}",
            "main",
        ],
        cwd=os.path.dirname(__file__),
    )



"""
Created on 28.11.2016

@author: freu_se
"""

from collections import OrderedDict

import numpy as np
import pandas as pd
from mpl_toolkits.mplot3d import Axes3D
from patme.service.logger import log
from patme.service.stringutils import indent

from delismm.model.customsystemfunction import BoundsHandler
from delismm.model.doe import FullFactorialDesign
from delismm.model.samplecalculator import getY


class ModelPlotter:
    """This class is supposed to plot models of various kind.

    reqired attributes of models::

    - callable with vector of params as input, the output is a list where the first item is used
    - lowerBounds, upperBounds
    - name
    - parameterNames

    from delismm.model.customsystemfunction import ExampleFunction, ExampleFunctionModified
    from delis.service.dataplotter import PlotGraphMatplotlib
    lb, ub = OrderedDict([('testParameter', 0.0),('testParameter2', 0.0)]), OrderedDict([('testParameter', 1.0),('testParameter2', 1.0)])
    models = [ExampleFunction(lb, ub), ExampleFunctionModified(lb, ub)]
    plotter = PlotGraphMatplotlib(shapeSubplots=(1,2))
    mp = ModelPlotter(models, plotter)
    mp.samplingSize = 101
    mp.plotModel([m.name for m in models], ['testParameter','testParameter2'], [0.5,0.5])
    plotter.activeNextSubplot()
    mp.plotModel([m.name for m in models], ['testParameter'], [0.5,0.5], [0.,0.], [0.5,1.0])
    # plotter.show()
    plotter.closeFigure()
    """

    def __init__(self, models, plotInsance):
        """doc"""
        self.samplingSize = 101
        """number of ticks in 1 direction to evaluate the function to plot"""
        self.plotter = plotInsance
        self._plotData = []
        """Cache for data to be plot until it is plot. Structure of each line: [funcOrAttr, args, kwargs]"""
        self._setModels(models)

    def _setModels(self, models):
        """is called on initialization or after a reset"""
        if len(models) == 0:
            return
        self._modelDict = OrderedDict()
        self._firstModel = models[0]
        self._modelDict[self._firstModel.name] = models[0]
        for model in models[1:]:
            self.addModel(model)

    def resetModels(self, models):
        """deletes all actual models and uses the given models"""
        self._setModels(models)

    def addModel(self, model):
        """doc"""
        lb, ub = self.lowerBounds, self.upperBounds
        if not np.allclose(lb, model.lowerBounds) or not np.allclose(ub, model.upperBounds):
            raise ValueError("No matching bounds of these models: " + str(model[0].name, ", " + str(model.name)))
        if model.name in self._modelDict:
            raise ValueError('There are models with equal names: "' + str(model.name) + '"')
        self._modelDict[model.name] = model

    def plotModel(
        self,
        modelNames,
        parameterNames,
        nominalValues,
        lowerBounds=None,
        upperBounds=None,
        title="",
        plotKind=None,
        samplingSizes=None,
        lineTrace3d=False,
        legendLocation="upper right",
    ):
        """

        :param lineTrace3d: For 3d contour plots, lines of both parameters are plot, that depict the trace on the surface in
            respect to the nominal value of the other parameter respectively
        :param legendLocation: string, defining the legends location. One of these: ['best', 'upper right', 'upper left', 'lower left', 'lower right', 'right', 'center left', 'center right', 'lower center', 'upper center', 'center']
        """
        self.collectPlotData(
            modelNames,
            parameterNames,
            nominalValues,
            lowerBounds,
            upperBounds,
            title,
            plotKind,
            samplingSizes,
            lineTrace3d,
            legendLocation,
        )
        self.plotCachedData()

    def collectPlotData(
        self,
        modelNames,
        parameterNames,
        nominalValues,
        lowerBounds=None,
        upperBounds=None,
        title="",
        plotKind=None,
        samplingSizes=None,
        lineTrace3d=False,
        legendLocation="upper right",
    ):
        """doc"""
        lowerBounds, upperBounds = self._checkPlotModelInputs(parameterNames, nominalValues, lowerBounds, upperBounds)

        if samplingSizes is None:
            samplingSizes = [self.samplingSize] * len(modelNames)

        if plotKind is None:
            plotKind = self._getPlotKind(parameterNames)

        if not all([samplingSize == samplingSizes[0] for samplingSize in samplingSizes]):
            # different sampling size among the models. This does not work in one data frame
            for modelName, samplingSize in zip(modelNames, samplingSizes):
                self.collectPlotData(
                    [modelName],
                    parameterNames,
                    nominalValues,
                    lowerBounds,
                    upperBounds,
                    title,
                    plotKind,
                    [samplingSize],
                    lineTrace3d,
                )
            return

        df = self._getPlotDataFrame(modelNames, parameterNames, nominalValues, lowerBounds, upperBounds, samplingSizes)
        self._plotData.append(
            [self.plotter.plot, [df], {"kind": plotKind, "title": title, "legendKwArgs": {"loc": legendLocation}}]
        )

        if lineTrace3d:
            if plotKind != "contour3d":
                log.warning('Will not create lineTrace3d since the plot kind is not "contour3d"')
                return
            if len(parameterNames) != 2:
                raise Exception("There must be excatly 2 parameters given for the lineTrace3d")
            allParameterNames = self._firstModel.parameterNames
            # first param
            df = self._getPlotDataFrame(
                modelNames, [parameterNames[0]], nominalValues, lowerBounds, upperBounds, samplingSizes
            )
            otherParameterNominalValue = nominalValues[allParameterNames.index(parameterNames[1])]

            self._plotData.append(
                [
                    self.plotter.plotLine3D,
                    (
                        df[parameterNames[0]],
                        [otherParameterNominalValue] * len(df[parameterNames[0]]),
                        df[modelNames[0]],
                    ),
                    {"legendKwArgs": {"loc": legendLocation}},
                ]
            )
            # second param
            df = self._getPlotDataFrame(
                modelNames, [parameterNames[1]], nominalValues, lowerBounds, upperBounds, samplingSizes
            )
            otherParameterNominalValue = nominalValues[allParameterNames.index(parameterNames[0])]
            self._plotData.append(
                [
                    self.plotter.plotLine3D,
                    [
                        [otherParameterNominalValue] * len(df[parameterNames[1]]),
                        df[parameterNames[1]],
                        df[modelNames[0]],
                    ],
                    {"legendKwArgs": {"loc": legendLocation}},
                ]
            )

    def plotCachedData(self):
        """doc"""
        for funcOrAttr, args, kwargs in self._plotData:
            if hasattr(funcOrAttr, "__call__"):
                funcOrAttr(*args, **kwargs)
            else:
                setattr(self.plotter, funcOrAttr, args[0])
        self._plotData = []

    def _getPlotKind(self, parameterNames):
        """doc"""
        if len(parameterNames) == 1:
            return "line"
        if isinstance(self.plotter.ax, Axes3D):
            return "contour3d"
        return "contour"

    def _getPlotDataFrame(self, modelNames, parameterNames, nominalValues, lowerBounds, upperBounds, samplingSizes):
        """Creates the data frame for each model by calling the model at a full factorial design(ffd) grid.

        A sampleX array is created, that contains nominals, where no parameter variation is done and
        the ffd values where the parameter is in parameterNames. Then, the model values are calculated and
        appended to the dataframe
        """
        allParameterNames = self._firstModel.parameterNames
        lowerBoundsPlot = [lowerBounds[allParameterNames.index(parameterName)] for parameterName in parameterNames]
        upperBoundsPlot = [upperBounds[allParameterNames.index(parameterName)] for parameterName in parameterNames]

        dfDict = OrderedDict()

        # sample y
        for modelName, samplingSize in zip(modelNames, samplingSizes):
            doe = FullFactorialDesign(samplingSize ** len(parameterNames), len(parameterNames))
            sampleXToInclude = BoundsHandler.scaleToBoundsStatic(
                doe.sampleXNormalized, lowerBoundsPlot, upperBoundsPlot
            )
            sampleX = self._getSampleX(parameterNames, nominalValues, dfDict, sampleXToInclude)
            if not modelName in self._modelDict:
                raise ValueError('Model with name "' + str(modelName) + '" does not exist.')
            model = self._modelDict[modelName]
            sampleY = getY(sampleX, model, verbose=False)
            dfDict[modelName] = [sample[0] for sample in sampleY]
            log.debug(
                "plot function [{}] with parameter(s) [{}], nominal(s) [{}] using these sampleX:\n{}\nand sampleY\n{}".format(
                    modelName, parameterNames, nominalValues, indent(sampleX, delim=" "), indent(sampleY, delim=" ")
                )
            )
        return pd.DataFrame(dfDict)

    def plotSurrogateSamples(
        self,
        modelNames,
        parameterNames,
        nominalValues,
        lowerBounds=None,
        upperBounds=None,
        title="",
        plotKind=None,
        sampleCorrThresholds=None,
        legendLocation="upper right",
    ):
        """the models with name modelNames must have attributes "sampleX" and "sampleY"!

        :param modelNames: list of names that must be included in the keys of self._modelDict
        :param parameterNames: list of parameter names that are used in each model
        :param nominalValues: list of floats defining the nominal values of unused parameters.
            It has the length of all parameters, so each parameter must have a nominal value
        :param sampleCorrThresholds: list of thresholds for the correlation values
            of the samples. Samples with a correlation value of less than this threshold will be omitted.
        """
        self.collectPlotDataSurrogateSamples(
            modelNames,
            parameterNames,
            nominalValues,
            lowerBounds,
            upperBounds,
            title,
            plotKind,
            sampleCorrThresholds,
            legendLocation,
        )
        self.plotCachedData()

    def collectPlotDataSurrogateSamples(
        self,
        modelNames,
        parameterNames,
        nominalValues,
        lowerBounds=None,
        upperBounds=None,
        title="",
        plotKind=None,
        sampleCorrThresholds=None,
        legendLocation="upper right",
    ):
        """doc"""
        if sampleCorrThresholds is None:
            sampleCorrThresholds = [0.0] * len(modelNames)
        for modelName, sampleCorrThreshold in zip(modelNames, sampleCorrThresholds):
            if plotKind is None:
                plotKind = self._getPlotKind(parameterNames)
            df = self._getPlotDataFrameForSurrogateSamples(
                modelName, parameterNames, nominalValues, lowerBounds, upperBounds, sampleCorrThreshold
            )
            if len(df) == 0:
                log.info("No samples with a correlation higher than " + str(sampleCorrThreshold))
                return
            log.debug("sample dataframe\n" + str(df))
            # plot samples
            usedDf = df
            if plotKind == "contour3d":
                usedPlotKind = "scatter3d"
            else:
                usedPlotKind = "scatter"
                if plotKind == "contour":
                    usedDf = df.drop(usedDf.columns[2:-1], axis=1)
            self._plotData.append(
                [
                    self.plotter.plot,
                    [usedDf],
                    {"kind": usedPlotKind, "title": title, "legendKwArgs": {"loc": legendLocation}},
                ]
            )
            # plot lines
            if plotKind == "line":
                allXAndZ = [([df.iloc[rowIndex, 0]] * 2, df.iloc[rowIndex, 1:3]) for rowIndex in range(len(df))]
                for x, z in allXAndZ:
                    self._plotData.append(["color", ["gray"], {}])
                    self._plotData.append(
                        [self.plotter.plotLine, [x, z], {"label": None, "legendKwArgs": {"loc": legendLocation}}]
                    )
            elif plotKind == "contour3d":
                allXAndZ = [
                    ([df.iloc[rowIndex, 0]] * 2, [df.iloc[rowIndex, 1]] * 2, df.iloc[rowIndex, 2:4])
                    for rowIndex in range(len(df))
                ]
                for x, y, z in allXAndZ:
                    self._plotData.append(["color", ["gray"], {}])
                    self._plotData.append(
                        [self.plotter.plotLine3D, [x, y, z], {"label": None, "legendKwArgs": {"loc": legendLocation}}]
                    )

    def _getPlotDataFrameForSurrogateSamples(
        self, modelName, parameterNames, nominalValues, lowerBounds, upperBounds, sampleCorrThreshold
    ):
        """Creates the data frame for each model by collecting the proper samples from the surrogate.

        A sampleX array is created, that contains nominals, where no parameter variation is done and
        the ffd values where the parameter is in parameterNames.
        """
        model = self._modelDict[modelName]
        allParameterNames = model.parameterNames
        dfDict = OrderedDict()
        lowerBoundsPlot, upperBoundsPlot = [], []

        # plot sample x
        for parameterName in parameterNames:
            parameterIndex = allParameterNames.index(parameterName)
            lowerBoundsPlot.append(lowerBounds[parameterIndex])
            upperBoundsPlot.append(upperBounds[parameterIndex])
            dfDict[parameterName] = model.sampleX[parameterIndex]

        # sample y projected
        plotSampleX = self._getSampleX(parameterNames, nominalValues, {}, np.array(list(dfDict.values())))

        sampleY = getY(plotSampleX, model, verbose=False)
        dfDict[modelName + "_project"] = [sample[0] for sample in sampleY]

        dfDict[modelName] = model.sampleY

        # calculate correlation values
        plotSampleXNormalized = BoundsHandler.scaleTo01Static(plotSampleX, self.lowerBounds, self.upperBounds)
        sampleXNormalized = BoundsHandler.scaleTo01Static(model.sampleX, self.lowerBounds, self.upperBounds)
        correlations = model.correlationFunctionNSamples(np.array(plotSampleXNormalized.T), sampleXNormalized.T)[0]
        dfDict["scatterColor"] = correlations

        df = pd.DataFrame(dfDict)

        # drop samples that are out of plot bounds
        lowerBoundsPlot, upperBoundsPlot = np.array(lowerBoundsPlot), np.array(upperBoundsPlot)
        inBounds = 1
        for parameterName, lowerBoundPlot, upperBoundPlot in zip(parameterNames, lowerBoundsPlot, upperBoundsPlot):
            inBounds &= (lowerBoundPlot - 1e-8 < df[parameterName]) & (df[parameterName] < upperBoundPlot + 1e-8)
        outOfBoundsIndex = inBounds[inBounds == False].index
        df.drop(outOfBoundsIndex, inplace=True)

        # drop samples that have a lower correlation value than the given threshold
        lowCorrelation = df["scatterColor"] < sampleCorrThreshold
        lowCorrelationIndex = lowCorrelation[lowCorrelation == True].index
        df.drop(lowCorrelationIndex, inplace=True)
        df.index = pd.RangeIndex(0, len(df), 1)

        # add two samples with correlation color zero and one in order to get the color of the plot range correct
        # they shold be plot first, in order to be replot by the real correlation
        if len(df):
            df = df.iloc[::-1]
            df = df.append(df.loc[0], ignore_index=True)  # copy last row, add it to df
            df = df.append(df.loc[len(df) - 1], ignore_index=True)  # copy last row, add it to df
            df = df.iloc[::-1]
            df["scatterColor"][len(df) - 2] = sampleCorrThreshold
            df["scatterColor"][len(df) - 1] = 1.0

        return df

    def _getSampleX(self, parameterNames, nominalValues, dfDict, sampleXToInclude):
        """Creates a sampleX array, that contains nominals, where no parameter variation is done and
        the values of sampleXToInclude where the parameter is in parameterNames."""
        allParameterNames = self._firstModel.parameterNames
        numberOfSamples = len(sampleXToInclude[0])
        sampleX = np.array([nominalValues] * numberOfSamples).T
        for parameterName, samples in zip(parameterNames, sampleXToInclude):
            parameterIndex = allParameterNames.index(parameterName)
            dfDict[parameterName] = samples
            sampleX[parameterIndex, :] = samples
        return sampleX

    def _checkPlotModelInputs(self, parameterNames, nominalValues, lowerBounds, upperBounds):
        """doc"""
        if len(nominalValues) != self.p:
            raise ValueError("Size of nominal values is wrong.")
        if lowerBounds is None:
            lowerBounds = self.lowerBounds
        elif len(lowerBounds) != self.p:
            raise ValueError("Size of lower bounds is wrong.")
        elif np.any(self.lowerBounds - np.abs(self.lowerBounds) * 0.02 > lowerBounds):
            raise ValueError("Given bounds are lower than the models lower bounds.")
        if upperBounds is None:
            upperBounds = self.upperBounds
        elif len(upperBounds) != self.p:
            raise ValueError("Size of upper bounds is wrong.")
        elif np.any(self.upperBounds + np.abs(self.upperBounds) * 0.02 < upperBounds):
            raise ValueError("Given bounds are greater than the models upper bounds.")
        if len(parameterNames) > 2:
            raise ValueError("Can only plot 2 parameters")
        return lowerBounds, upperBounds

    def _getLowerBounds(self):
        """doc"""
        return self._firstModel.lowerBounds

    def _getUpperBounds(self):
        """doc"""
        return self._firstModel.upperBounds

    def _getParameterCount(self):
        """doc"""
        return len(self.lowerBounds)

    p = property(fget=_getParameterCount)
    lowerBounds = property(fget=_getLowerBounds)
    upperBounds = property(fget=_getUpperBounds)


if __name__ == "__main__":
    import doctest

    doctest.testmod()
#     from delismm.model.customsystemfunction import ExampleFunction, ExampleFunctionModified
#     from delis.service.dataplotter import PlotGraphMatplotlib
#     lb, ub = OrderedDict([('testParameter', 0.0),('testParameter2', 0.0)]), OrderedDict([('testParameter', 1.0),('testParameter2', 1.0)])
#     models = [ExampleFunction(lb, ub)]#, ExampleFunctionModified(lb, ub)]
#     plotter = PlotGraphMatplotlib(shapeSubplots=(2,2), is3dAxes=[False,True,False,False])
#     mp = ModelPlotter(models, plotter)
#     mp.samplingSize = 101
#     mp.plotModel([m.name for m in models], ['testParameter','testParameter2'], [0.5,0.5], [0.,0.], [0.65,0.65])
#     plotter.activeNextSubplot()
#     mp.plotModel([m.name for m in models], ['testParameter','testParameter2'], [0.5,0.5], [0.,0.], [0.65,0.65])
#     plotter.activeNextSubplot()
#     mp.plotModel([m.name for m in models], ['testParameter'], [0.5,0.5], [0.,0.], [0.5,1.0])
#     plotter.show()
#     plotter.closeFigure()



"""
This module is inteded to hande sampleX and sampleY values that may be reduced by sampleX/sampleY-columns and based on some criteria of the samples.

Created on 16.10.2017

@author: freu_se
"""

from collections import OrderedDict

import numpy as np
import pandas as pd


class SampleSetReducer:
    """classdoc"""

    def __init__(self, parameterNames, resultNames, nameToEqCriterionTuples=None, reduceParameterResultNames=None):
        """Initialization of SampleSetReducer

        Caution!! SampleX and SampleY are transposed to each other! sampleX.T matches sampleY

        :param parameterNames: list with names of the input parameters
        :param resultNames: list with names of the output parameters
        :param nameToEqCriterionTuples: list with tuples that define a input/output parameter and a value as equality criterion
            in order to remove rows from the dataset. (see test for an example)
        :param reduceParameterResultNames: list with input and/or output parameter names that should be removed
        """
        self.parameterNames = parameterNames
        self.resultNames = resultNames
        self.reduceParameterResultNames = reduceParameterResultNames
        self.nameToEqCriterionTuples = nameToEqCriterionTuples

    def getReducedSamples(self, sampleX, sampleY):
        """Method that reduces the given sampleX and sampleY"""
        # =======================================================================
        # setup dataframe
        # =======================================================================
        oDictInput = [(name, sampleColumn) for name, sampleColumn in zip(self.parameterNames, sampleX)]
        oDictInput += [(name, sampleColumn) for name, sampleColumn in zip(self.resultNames, zip(*sampleY))]
        samplesDf = pd.DataFrame(OrderedDict(oDictInput))

        # =======================================================================
        # use only row fulfilling the given criterion
        # =======================================================================
        if self.nameToEqCriterionTuples is not None:
            for name, criterion in self.nameToEqCriterionTuples:
                if not name in samplesDf.columns:
                    return sampleX, sampleY
                samplesDf = samplesDf[samplesDf[name] == criterion]

        # =======================================================================
        # reduce sample columns
        # =======================================================================
        parameterNames = self.parameterNames
        resultNames = self.resultNames
        if self.reduceParameterResultNames is not None:
            parameterNames = [
                parameterName
                for parameterName in parameterNames
                if parameterName not in self.reduceParameterResultNames
            ]
            resultNames = [
                resultName for resultName in resultNames if resultName not in self.reduceParameterResultNames
            ]
            newColumns = parameterNames + resultNames
            samplesDf = samplesDf.loc[:, newColumns]

        # =======================================================================
        # create result sample collections
        # =======================================================================
        sampleX = np.array(samplesDf.loc[:, parameterNames])
        sampleY = samplesDf.loc[:, resultNames].values.tolist()

        return sampleX.T, sampleY



# -*- coding:utf-8 -*-
# Copyright (C) 2013 Deutsches Zentrum fuer Luft- und Raumfahrt(DLR, German Aerospace Center) <www.dlr.de>
"""
Created on 28.09.2011

"""

import copy
import os
import pickle
import random
import time
from collections import OrderedDict

import numpy as np
from patme.service.logger import log
from patme.service.stringutils import indent
from scipy import linalg, special
from scipy.optimize import differential_evolution, minimize

from delismm import version as delismmVersion
from delismm.model.apimodels import SurrogateInfo
from delismm.model.customsystemfunction import (
    BoundsHandler,
    CrossValidationValuesCalculatorKriging,
    ThetaOptimizerKriging,
)
from delismm.model.samplecalculator import getY
from delismm.service.exception import DelisMMError


class AbstractSurrogate:
    """Abstract class with interface specifications for surrogate methods"""

    identicSampleXThreshold = 1e-8
    """Threshold for self.removeIdenticSampleX"""

    def __init__(self, sampleX, sampleY, lowerBounds, upperBounds, parameterNames, resultNames):
        """Initialization of surrogate samples

        :param sampleX: numpy array of shape [p,n] (p: number of parameters; n: number of samples)
        :param sampleY: numpy array of shape [n] or [n, x] (x is arbitrary - only x[0] is used)
        :param lowerBounds: numpy array with lower bounds of sample X.
            If None, lowerBounds are calculated from sampleX
        :param upperBounds: numpy array with upper bounds of sample X.
            If None, upperBounds are calculated from sampleX
        :param parameterNames: List with names of each parameter. If None, [p_i] will be used
        :param resultNames: list with the names of the resulting values
        """

        # PARAMETER section ###################################################
        if not isinstance(sampleX, np.ndarray):
            sampleX = np.array(sampleX)
        if lowerBounds is None:
            lowerBounds = np.zeros((sampleX.shape[0],))
        if upperBounds is None:
            upperBounds = np.ones((sampleX.shape[0],))
        self.sampleXNormalized = BoundsHandler.scaleTo01Static(sampleX, lowerBounds, upperBounds)
        self.lowerBounds = lowerBounds
        self.upperBounds = upperBounds
        if parameterNames is None:
            parameterNames = ["p" + str(parameterIndex) for parameterIndex in range(len(self.lowerBounds))]
        self.parameterNames = parameterNames
        self.log10SampleX = np.array([False] * len(parameterNames))
        """Flags if the given sampleX for each parameter are the log10 of the original samples.
        If yes, __call__ will calculate log10 for the respective parameters"""

        # RESULT section ######################################################
        if resultNames is None:
            resultNames = ["resultName"]
        self.resultNames = resultNames
        try:
            sampleY = np.array(sampleY, dtype=np.float64)
        except ValueError:
            # case if sampleY is np.array with ndim==1 but each item contains a list. This happens if there are strings in the previous array
            sampleY = np.array([it[0] for it in sampleY])
        if sampleY.ndim == 2:
            sampleY = sampleY[:, 0]
        elif sampleY.ndim > 2:
            raise DelisMMError("Expected an array of dimension 1 or 2 but got {}".format(sampleY.ndim))
        self.sampleY = np.array(sampleY, dtype=np.float64)
        self.log10SampleY = False
        """Flag if the given sampleY are the log10 of the original samples.
        If yes, __call__ will return power(10, logY)"""
        self.logLevel = log.INFO
        """The loglevel might be changed in some output situations in order to reduce the output"""
        self.name = "surrogate"
        self.removeIdenticSampleX()

    def createSurrogateModel(self):
        """Method that creates the surrogate model. Surrogate settings may be done
        prior to this call"""
        raise NotImplementedError("This method must be implemented in a subclass")

    def targetFunction(self, pointOfInterest):
        """
        Method that evaluates the model on a specific point
        :param pointOfInterest: point to be evaluated
        :return: model return
        """
        raise NotImplementedError("This method must be implemented in a subclass")

    def removeIdenticSampleX(self):
        """Removes samples that identical in respect to a given threshold"""
        newSampleX = []
        newSampleY = []
        for i, samples in enumerate(zip(self.sampleXNormalized.T, self.sampleY)):
            sample1, sampleY = samples
            hasIdenticSample = False
            for j, sample2 in enumerate(self.sampleXNormalized.T[i + 1 :]):
                if np.allclose(sample1, sample2, self.identicSampleXThreshold):
                    hasIdenticSample = True
                    break
            if hasIdenticSample:
                log.warning(
                    "Sample at index {} and {} (zero based counting) are identical. I will remove the first one.".format(
                        i, j + 1 + i
                    )
                )
            else:
                newSampleX.append(sample1)
                newSampleY.append(sampleY)

        if len(newSampleX) != self.sampleXNormalized.shape[1]:
            self.sampleXNormalized = np.array(newSampleX).T
            self.sampleY = np.array(newSampleY)

    def save(self, filename, path=""):
        """saves the surrogate to the file in filename."""
        if path and not os.path.exists(path):
            os.makedirs(path)
        if path:
            filename = os.path.join(path, filename)
        with open(filename, "wb") as f:
            pickle.dump(self, f)
        log.info("Surrogate saved at " + filename)

    @staticmethod
    def load(filename, path=""):
        """loads a surrogate from the file in filename."""
        if path:
            filename = os.path.join(path, filename)
        if not os.path.exists(filename):
            raise DelisMMError("No such file: " + filename)
        with open(filename, "rb") as f:
            surrogate = pickle.load(f)
        if isinstance(surrogate, AbstractKriging) and not surrogate.isCreated:
            surrogate.createSurrogateModel(surrogate.theta)
        surrogate.isCorrectVersion()
        if hasattr(surrogate, "log10SampleX"):
            # probably due to loading pickled surrogates, the bool values changed. Thus setting actual bools
            # print(surrogate.log10SampleX[0] is False)
            surrogate.log10SampleX = [(False if z == False else True) for z in surrogate.log10SampleX]
            # print(surrogate.log10SampleX[0] is False)
        log.info('Surrogate with name "{}" loaded from {}'.format(surrogate.name, filename))
        return surrogate

    def isCorrectVersion(self):
        """Checks the version of a surrogate instance with the actual Surrogate class version

        This method must be implemented in a subclass in order to take effect."""
        return True

    def _getSampleX(self):
        """doc"""
        return BoundsHandler.scaleToBoundsStatic(self.sampleXNormalized, self.lowerBounds, self.upperBounds)

    sampleX = property(fget=_getSampleX)


class ImproveKrigingModel(AbstractSurrogate):
    """
    This is a class which contains methods to improve an existing Kriging Model. Therefore the class AbstractKriging inherits the methods.
    You should not instanciate this class directly.
    """

    def refineKrigingModelNoGridding(
        self,
        targetFunctionInstance,
        maxNrOfNewPoints,
        maxTime=60 * 60 * 24 * 2,
        initOptiPoints=None,
        samplingOptCrit="mse",
        doRemoteCall=False,
        runDir=None,
    ):
        """Refines the kriging model without using adaptive gridding. This might be good for initial
        new samples that should be spread over the design space without checking maxRelError.
        A new theta value is calculated at the end of this method

        :param targetFunctionInstance: instance of a delismm.model.customsystemfunction class (or a derived one),
            that can be evaluated to generate new sample data
        :param maxNrOfNewPoints: number of new sample points
        :param maxTime: time limit [s]
        :param initOptiPoints: number of restarts of local optimization
        :param samplingOptCrit: string that defines the function to maximize: ['mse','det','vfmError','vfmCorrelation'].
        :return: new Kriging model
        """
        if initOptiPoints is None:
            initOptiPoints = 5 * self.p
        log.info("Refine Kriging model without adaptive gridding")
        log.info(
            "samplingOptCrit = "
            + str(samplingOptCrit)
            + ", maxTime = "
            + str(maxTime)
            + ", nrOfNewPoints = "
            + str(maxNrOfNewPoints)
        )
        startTime = time.time()
        lower = np.array([0.0] * self.p)
        upper = np.array([1.0] * self.p)
        krigingModel = copy.copy(self)
        newXs = []
        for _ in range(maxNrOfNewPoints):
            if time.time() - startTime > maxTime:
                log.info("Time Limit (" + str(maxTime) + ") has been reached!")
                break
            newX = krigingModel.newX(lower, upper, None, initOptiPoints, samplingOptCrit)
            newXs.append(newX)
            newY = targetFunctionInstance(BoundsHandler.scaleToBoundsStatic(newX, self.lowerBounds, self.upperBounds))
            if not isinstance(newY, str) and hasattr(newY, "__iter__"):
                newY = newY[0]
            nKrigingModel = krigingModel.n
            sampleXNormalized = np.zeros((self.p, nKrigingModel + 1))
            sampleXNormalized[:, :nKrigingModel] = copy.copy(krigingModel.sampleXNormalized)
            sampleXNormalized[:, -1] = newX
            sampleY = np.zeros((nKrigingModel + 1,))
            sampleY[:nKrigingModel] = copy.copy(krigingModel.sampleY)
            sampleY[-1] = newY
            krigingModel = self.getCopyNewSamples(sampleXNormalized, sampleY)
            oldLogLevel = krigingModel.logLevel
            krigingModel.logLevel = log.DEBUG
            krigingModel.createSurrogateModel()
            krigingModel.logLevel = oldLogLevel
        # recalculate theta
        krigingModel.createSurrogateModel(self.theta, forceOptTheta=True, doRemoteCall=doRemoteCall, runDir=runDir)
        log.info("I found new sample points: " + indent(newXs))
        log.info("Kriging Model refinement finished. Added " + str(krigingModel.n - self.n) + " sample points.")
        return krigingModel

    def refineKrigingModel(
        self,
        targetFunctionInstance,
        maxRelError=0.05,
        maxLoops=float("inf"),
        maxTime=60 * 60 * 24 * 2,
        maxNumberNewGriddingPoints=50,
        initOptiPoints=None,
        samplingOptCrit="mse",
        doRemoteCall=False,
        runDir=None,
    ):
        """Creates a new Kriging model by adaptive gridding.
        :param targetFunctionInstance: instance of a delismm.model.customsystemfunction class (or a derived one),
            that can be evaluated to generate new sample data
        :param maxLoops: maximal number of loops
        :param maxTime: time limit [s]
        :param maxNumberNewGriddingPoints: maximal number of new sample points
        :param maxRelError: target accuracy, evaluation criterion for cross validation
        :param initOptiPoints: number of restarts of local optimization
        :param samplingOptCrit: string that defines the function to maximize: ['mse','det','vfmError','vfmCorrelation'].
        :return: new Kriging model
        """
        if initOptiPoints is None:
            initOptiPoints = 5 * self.p
        log.info(
            "samplingOptCrit = "
            + str(samplingOptCrit)
            + ", maxRelError = "
            + str(maxRelError)
            + ", maxTime = "
            + str(maxTime)
            + ", maxNumberNewGriddingPoints = "
            + str(maxNumberNewGriddingPoints)
        )
        loopCount = 1
        startTime = time.time()
        krigingModel = copy.copy(self)
        while True:
            if loopCount > maxLoops:
                log.info("Maximal number of loops (" + str(maxLoops) + ") has been reached!")
                break
            if krigingModel.n - self.n > maxNumberNewGriddingPoints:
                log.info("Maximal number of new points (" + str(maxNumberNewGriddingPoints) + ") has been reached!")
                break
            if time.time() - startTime > maxTime:
                log.info("Time Limit (" + str(maxTime) + ") has been reached!")
                break
            log.info("Refine Kriging model. Loop " + str(loopCount))

            oldLogLevel = krigingModel.logLevel
            krigingModel.logLevel = log.DEBUG
            newSampleXNormalized = krigingModel._adaptiveGridding(
                maxRelError, self.n + maxNumberNewGriddingPoints - krigingModel.n, initOptiPoints, samplingOptCrit
            )
            krigingModel.logLevel = oldLogLevel

            if newSampleXNormalized is None:
                log.info("No new Samples found with maxRelError (" + str(maxRelError) + "). Refinement finished.")
                break
            newSampleCount = newSampleXNormalized.shape[1] + krigingModel.n
            oldSampleXNormalized = copy.copy(krigingModel.sampleXNormalized)
            ssampleY = copy.copy(krigingModel.sampleY)

            # calculate new sampleY
            newSampleX = BoundsHandler.scaleToBoundsStatic(newSampleXNormalized, self.lowerBounds, self.upperBounds)
            newSampleY = np.array(getY(newSampleX, targetFunctionInstance)).flatten()
            # put samples thogether
            sampleXNormalized = np.zeros((self.p, newSampleCount))
            sampleXNormalized[:, 0 : krigingModel.n] = oldSampleXNormalized
            sampleXNormalized[:, krigingModel.n : newSampleCount] = newSampleXNormalized
            sampleY = np.zeros((newSampleCount,))
            sampleY[0 : krigingModel.n] = ssampleY
            sampleY[krigingModel.n : newSampleCount] = newSampleY
            krigingModel = self.getCopyNewSamples(sampleXNormalized, sampleY)
            # recalculate theta
            krigingModel.createSurrogateModel(self.theta, forceOptTheta=True, doRemoteCall=doRemoteCall, runDir=runDir)
            loopCount += 1
        log.info("Kriging Model refinement finished. Added " + str(krigingModel.n - self.n) + " sample points.")
        return krigingModel

    def _adaptiveGridding(self, maxRelError, maxNumberNewGriddingPoints, initOptiPoints, samplingOptCrit):
        """Splits the design space in equal-sized, hypercuboid cells and evaluates the cells via cross validation.
        Returns new sample points in 'bad' cells.
        :param maxRelError: target relative accuracy (relative to the associated sampleY-value)
        :param maxNumberNewGriddingPoints: maximal number of new points that can be selected
        :param initOptiPoints: number of initial points for local optimization
        :param samplingOptCrit: string that defines the function to maximize: ['mse','det','vfmError','vfmCorrelation'].
        :return: numpy-matrix with new sample points (column by column)

        """
        maxDivisions = 10  # maximal number of interval divisions
        minDivisions = 2  # minimal number of interval divisions

        # create a tuple divisionNumbers in which the number of interval divisions of each koordinate axis is stated
        # create a np.array intervalLen in which the length of the subintervals is stated
        divisionNumbers = []
        intervalLen = []
        for thetaJ in self.theta:
            length = 1.0
            if self.corr == "cubic":
                thetaJ = thetaJ * 10  # cubic correlation only allows theta to be at max = 1
            number = int(length * thetaJ) + 1
            if number < minDivisions:
                number = minDivisions
            if number > maxDivisions:
                number = maxDivisions
            intervalLen.append(length / number)
            divisionNumbers.append(number)
        intervalLen = np.array(intervalLen)
        divisionNumbers = tuple(divisionNumbers)
        # list of all adapted sample points
        newXs = []
        # iterate over all sub-hypercubes defined by the subintervals
        for i in range(np.prod(divisionNumbers)):
            if maxNumberNewGriddingPoints - len(newXs) <= 0:
                break
            # lower and upper bounds of the hypercube
            lower = np.array(np.unravel_index(i, divisionNumbers)) * intervalLen
            upper = np.ones((self.p,)) * intervalLen + lower
            # copy of the sampleXNormalized array as a list; the entries of the list are the sample points as np.arrays
            subset = copy.copy(list(self.sampleXNormalized.T))
            # iterate over the different coordinate axes and over the sample points and test, whether a point lies within the hypercube
            for k in range(self.p):
                pointIndex = 0
                while pointIndex < len(subset):
                    # if the point doesn't lie within the hypercube, remove it from subset
                    if (
                        subset[pointIndex][k] < lower[k]
                        or subset[pointIndex][k] > upper[k]
                        or subset[pointIndex][k] < 1e-15
                        or 1 - subset[pointIndex][k] < 1e-15
                    ):
                        subset.pop(pointIndex)
                    else:
                        pointIndex = pointIndex + 1
            # do a cross validation for subset
            if len(subset) == 0:
                badIndex, badValue = None, 0.0
            else:
                badIndex, badValue = self._crossValidationSubset(subset)[:-1]
            if badValue > np.abs(np.max(self.sampleY) - np.min(self.sampleY)) * maxRelError or badIndex is None:
                newX = self.newX(lower, upper, badIndex, initOptiPoints, samplingOptCrit)
                isNew = True
                for k in range(len(newXs)):
                    if np.linalg.norm(newX - newXs[k]) < 1e-7:
                        isNew = False
                        break
                if isNew:
                    newXs.append(newX)
        if newXs == []:
            log.debug("No new sample points required.")
            return None
        log.info("I found new sample points: " + indent(newXs))
        return np.array(newXs).T

    def getCrossValidationValues(self, doRemoteCall=False, runDir=None):
        """Performs a crossvalidation of this metamodel and returns the cross validation values

        :return: np.ndarray of len self.n with values of crossValidation in the
            order like self.sampleXNormalized
        """
        oldLogLevel = self.logLevel
        self.logLevel = log.DEBUG
        crossValidationDifferences = self._crossValidationSubset(
            self.sampleXNormalized.T, doRemoteCall=doRemoteCall, runDir=runDir
        )[2]
        crossValidationDifferences = np.array(crossValidationDifferences).flatten()
        self.logLevel = oldLogLevel
        return self.sampleY - crossValidationDifferences

    def _crossValidationSubset(self, subset, doParallelization=True, doRemoteCall=False, runDir=None):
        """Returns the maximum of the cross validation of subset's elements.
        It is a measure for the error of the metamodel.

        :param subset: set of points for cross validation, must be a subset of support points sampleXNormalized; type = list
        :return: triple (index of the worst cross-validation value, cross-validation difference at that index, np array with all cross-validation values),
            where index is the index of the maximum cross-validation point in sampleXNormalized
            and cross-validation value is the according value of cross validation.
        """
        if len(subset) == 0:
            return (-1, 10**20, [])
        listOfSupports = list(self.sampleXNormalized.T)
        listOfSupportValues = list(self.sampleY)
        # Erstellen einer Liste mit cross validation values
        args = (listOfSupports, listOfSupportValues)
        if 0:  # local serial calculation or parallel calculation
            indexInSampleX = []
            crossValidationDiffs = []
            for crossX in subset:
                crossValidationDiff, index = self.crossValidation(crossX, *args)
                crossValidationDiffs.append(crossValidationDiff)
                indexInSampleX.append(index)
        else:
            cvTargetFunction = CrossValidationValuesCalculatorKriging(self, args, doRemoteCall, runDir=runDir)
            if not doParallelization:
                cvTargetFunction.doParallelization = []
            log.info('Perform crossvalidation for kriging model "{}"'.format(self.name))
            result = getY(np.array(subset).T, cvTargetFunction, False)
            crossValidationDiffs, indexInSampleX = list(zip(*result))
        maxi = np.max(np.abs(crossValidationDiffs))
        maxiIndexInSubset = np.abs(crossValidationDiffs).argmax()
        maxiIndexInSampleX = indexInSampleX[maxiIndexInSubset]
        absRelError = maxi / np.abs(np.max(self.sampleY) - np.min(self.sampleY))
        log.info(
            "Max abs cross validation abs and scaled error : "
            + str(maxi)
            + ", "
            + str(absRelError)
            + " at "
            + str(subset[maxiIndexInSubset])
        )

        return (maxiIndexInSampleX, maxi, crossValidationDiffs)

    def crossValidation(self, crossX, listOfSupports, listOfSupportValues):
        """Calculates the cross validation diff of one sample"""
        supports = copy.copy(listOfSupports)
        supportValues = copy.copy(listOfSupportValues)
        # Entferne eine Stuetzstelle aus der Liste
        index = 0
        # find correct index for crossX
        while index < len(supports) and not np.linalg.norm(crossX - supports[index]) < 1.0e-15:
            index = index + 1
        if index == len(supports):
            raise ValueError("sample of subset not found in self.sampleXNormalized")
        sampleY = supportValues[index]
        supports.pop(index)
        supportValues.pop(index)
        supports = np.array(supports).T
        # Metamodell ueber die restlichen Stuetzstellen
        crossKriging = self.getCopyNewSamples(supports, np.array(supportValues))
        crossKriging.createSurrogateModel(theta=self.theta)
        # fuege Differenz von original Metamodell-Wert und neuem Metamodell-Wert zu Liste hinzu
        crossY = crossKriging.callNormalized(crossX.T)
        if self.log10SampleY:
            # calculate in log-space
            crossY = np.log10(crossY)
        crossValidationDiff = sampleY - crossY
        return crossValidationDiff, index

    def newX(self, lower, upper, badIndex, initOptiPoints, samplingOptCrit):
        """Returns a np.array with the coordinates of a new point within the subcube [lower,upper].
        The new point is choosen via the method given in samplingOptCrit (by default: samplingOptCrit = 'mse')
        :param lower: array with lower bounds of the hypercube
        :param upper: array with upper bounds of the hypercube
        :param badIndex: index of the worst point of the subset
        :param initOptiPoints: number of start points for local optimization
        :param samplingOptCrit: string that defines the function to maximize: ['mse','det','vfmError','vfmCorrelation'].
        :return: the new point
        """
        # ===========================================================================================
        # create start values for local opt
        # ===========================================================================================
        initOptiPoints = 4
        bounds = []
        for i in range(self.p):
            bounds.append((lower[i], upper[i]))
        x0s = []
        if badIndex is None:  # no samples in the subset. Thus a sample should be created here
            x0s.append(lower + (upper - lower) / 2)
        else:
            badPlusEps = (
                1.0e-7 * np.ones(self.sampleXNormalized[:, badIndex].shape) + self.sampleXNormalized[:, badIndex]
            )
            badMinusEps = (
                -1.0e-7 * np.ones(self.sampleXNormalized[:, badIndex].shape) + self.sampleXNormalized[:, badIndex]
            )
            if not np.any(badPlusEps < lower) or np.any(badPlusEps > upper):
                # badPlusEps is within bounds
                x0s.append(list(badPlusEps))
            if not np.any(badMinusEps < lower) or np.any(badMinusEps > upper):
                # badMinusEps is within bounds
                x0s.append(list(badMinusEps))
        while len(x0s) <= initOptiPoints:
            newx0s = []
            for i in range(self.p):
                newx0s.append(random.uniform(lower[i], upper[i]))
            isIn = False
            for i in range(len(x0s)):
                if np.linalg.norm(np.array(x0s[i]) - np.array(newx0s), 2) <= 1.0e-15:
                    isIn = True
                    break
            if not isIn:
                x0s.append(newx0s)

        # ===========================================================================================
        # set opt function
        # ===========================================================================================
        f = self.getSamplingOptFunction(samplingOptCrit)
        fMin = lambda x: -1 * f(x)
        #         maxF0 = np.max([f(np.array(x).T) for x in x0s])
        #         f = self.getSamplingOptFunction(samplingOptCrit)
        #         fScaled = lambda x: -1 * f(x) / maxF0

        # =======================================================================
        # parform optimization
        # =======================================================================
        bounds = list(zip(lower, upper))
        #         optimum = self._getGlobalOptimum(f, bounds)
        optimum = self._doGlobalOpt(fMin, bounds, doRemoteCall=False)

        log.debug(
            "Opt results (TargetValue, thetas):  \n"
            + indent([[self.optThetaGlobalLocalType] + [optimum.fun] + list(optimum.x)])
        )
        return optimum.x

    def getSamplingOptFunction(self, samplingOptCrit):
        """Returns the function to optimize the location of a new sample point.

        :param samplingOptCrit: string that defines the function to maximize: ['mse','det','vfmError'].
        :return: function pointer to target function
        """
        if samplingOptCrit == "mse":
            return self.getMSEofPredictor
        elif samplingOptCrit == "det":
            return self._optimizeNewR
        elif samplingOptCrit == "vfmError":
            if not isinstance(self, HierarchicalKriging):
                raise DelisMMError("Could not use the relative hierarchical kriging error")
            # MSE * abs(relative error - 1)
            return self._optimizeVfmError
        elif samplingOptCrit == "vfmCorrelation":
            if not isinstance(self, HierarchicalKriging):
                raise DelisMMError("Could not use the relative hierarchical kriging error")
            # MSE * abs(Kriging_correlation)
            return self._optimizeMseAndCorrelation
        elif hasattr(samplingOptCrit, "__call__"):
            # samplingOptCrit is a kriging model or an other targetFunction
            # using the relative error function of self and the given function
            self.otherTargetFunc = samplingOptCrit
            return self._optimizeRelErrorWithOtherTargetFunc
        else:
            raise DelisMMError(
                "No proper criterion as function to maximize for new sample given. These are the allowed ones: ['mse','det','vfmError']"
            )

    def _optimizeRelErrorWithOtherTargetFunc(self, x):
        """doc"""
        valueSelf, valueOther = self(x), self.otherTargetFunc(x)
        return 1 - np.min([valueSelf, valueOther]) / np.max([valueSelf, valueOther])

    def _optimizeVfmError(self, x):
        """Only works for hierarchical models"""
        minByMax = np.min([self(x), self.lfKriging(x)]) / np.max([self(x), self.lfKriging(x)])
        return self.getMSEofPredictor(x) * (1 - minByMax)

    def _optimizeMseAndCorrelation(self, x):
        """Only works for hierarchical models"""
        corrVec = self.getCorrelationVector(x)
        correlationValue = (corrVec.T @ self.v)[0, 0]
        return self.getMSEofPredictor(x) * np.abs(correlationValue)

    def _optimizeNewR(self, newX):
        """Function for optimization process. This function needs to be minimized for optimal new sample point
        :param newX: point within subcube
        :return: ln(det(newR))
        """
        newR = np.zeros((self.n + 1, self.n + 1))
        newR[0 : self.n, 0 : self.n] = self._R
        for i in range(self.n):
            newR[self.n, i] = self._correlationFunction(self.sampleXNormalized[:, i], newX, self.theta)
            newR[i, self.n] = newR[self.n, i]
            newR[i, i] = 1.0e-11 + newR[i, i]
        newR[self.n, self.n] = 1.0e-11 + 1.0
        try:
            cholR = np.linalg.cholesky(newR)
        except np.linalg.LinAlgError:
            log.info("Matrix ist nicht positiv definit!")
            return -(10**20)
        lnDet = 0.0
        for i in range(self.n + 1):
            lnDet = lnDet + 2 * np.log(cholR[i, i])
        return lnDet


class AbstractKriging(ImproveKrigingModel):
    """init"""

    version = delismmVersion

    def __init__(
        self, sampleX, sampleY, lowerBounds, upperBounds, corr, regress="const", parameterNames=None, resultNames=None
    ):
        """Kriging class for common methods of Kriging and hierarchical Kriging"""
        AbstractSurrogate.__init__(self, sampleX, sampleY, lowerBounds, upperBounds, parameterNames, resultNames)

        self.version = self.version
        """set class attribute "version" as instance attribute for checking when the surrogate is loaded"""
        # =======================================================================
        # Kriging important public attributes
        # =======================================================================
        corrModels = ["gauss", "exp", "cubic"]
        if corr not in corrModels:
            raise DelisMMError(f"Wrong correlation model name given. These are applicable models: {corrModels}")
        self.corr = corr
        """ corr is a string which declares the used correlation model. corr can be 'gauss', for gaussian correlation,
            'exp', for exponential correlation or 'cubic' for cubic correlation. """

        regressModels = ["const", "linear"]
        if regress not in regressModels:
            raise DelisMMError(f"Wrong regression model name given. These are applicable models: {regressModels}")
        self.regress = regress
        """Regression model - either constant or linear regression"""

        self.n = self.sampleXNormalized.shape[1]  # n - Number of sample points
        """ n is the number of sample points """
        self.p = self.sampleXNormalized.shape[0]  # p - Number of design variables
        """ p is the dimension of the design space/ the number of model parameters """
        self.theta = None
        """ theta is the correlation parameter and defines the regression width of the regression function.
        It can be set by the keyword argument 'theta'. If it is not set by the user, theta is determined by
        an maximum likelihood optimization. It is a 1D array: np.array([theta_1,theta_2 .. theta_p])"""

        self.regularizationParameter = 1e-11
        """"regularization to the correlation matrix"""
        self.doRegularizationParameterOpt = True
        """Switch that performs the optimization of the regularization parameter"""

        self.name = "kriging"
        if self.resultNames:
            self.name += "_" + self.resultNames[0]

        # =======================================================================
        # theta optimization attributes
        # =======================================================================
        self.optThetaGlobalLocalType = "global"
        """Type of optimization run. These string values are valid: [global, local].
        Global or pure local optimization can be performed. After each global optimization run,
        the scipy optimizer performs a local optimization. The global optimizations are
        repeated self.optThetaGlobalAttempts times.
        """
        self.optThetaGlobalAttempts = 6
        """Integer, describing the attempts to perform a global local optimization"""
        self.optThetaMaxIter = 10**3
        """Maximal Iterations of one population within the theta optimization"""
        self.optThetaPrintout = False
        """print text output every <optThetaPrintout> iteration"""

        # =======================================================================
        # result values of a created model
        # =======================================================================
        self.correlationMatrixAlgo = "fast"
        """[fast, memory] String that determines the correlation matrix method used.
        _getCorrelationMatrixMemory is slow but requires less memory, _getCorrelationMatrixFast is fast but requires much memory"""
        self._R = None
        self._modifiedCorrMatrix = None
        """block matrix, eq 3.20
        [ R    F
          F.T  0 ]"""
        self.condition = None
        """ Condition number (2-norm) of modified correlation matrix (left hand side of system which must be solved to evaluate metamodell at a given point) """
        self.beta0 = None
        """ Regression parameters """
        self.v = None
        self._F = None
        """regressionFunctionValues F(n,p+1) = [f(x^1), ..., f(x^n)]"""

    def createSurrogateModel(self, theta=None, forceOptTheta=False, doRemoteCall=False, runDir=None):
        """Creates the kriging model. Here all result variables are set as private class attributes."""

        log.log(self.logLevel, "Start constructing a surrogate of type " + str(self.__class__.__name__))
        log.debug('Correlation model "%s"' % self.corr)
        if hasattr(self, "regress"):
            log.debug('Regression model "%s"' % self.regress)
        self._F = self._getRegressionFunctionValues()
        # Theta optimization
        if (theta is None and self.theta is None) or forceOptTheta:
            log.debug("Start theta optimization.")
            if not forceOptTheta and (theta is None and self.theta is not None):
                theta = self.theta
            self.theta, self.regularizationParameter = self._optimizeHyperparams(theta, doRemoteCall, runDir)
            log.log(self.logLevel, "Found theta. Optimization was successful.")
        else:
            if theta is not None:
                self.theta = np.abs(theta)
        log.debug("theta = " + str(self.theta))

        self._R = self._getCorrelationMatrix(self.theta, self.regularizationParameter)
        self._modifiedCorrMatrix = self._getModifiedCorrMatrix()
        """necessary for getMSEofPredictor"""
        self.condition = np.linalg.cond(self._modifiedCorrMatrix)
        """ Condition number (2-norm) of modified correlation matrix (left hand side of system which must be solved to evaluate metamodell at a given point) """
        log.debug("Condition number: " + str(self.condition))

        self.beta0 = self.getBeta(self._F, self._R, self.sampleY, symPos=False)
        """ Regression parameters """
        self.v = self.getV(self._F, self._R, self.sampleY, self.beta0)
        log.log(self.logLevel, "End constructing a surrogate of type " + str(self.__class__.__name__))

    def regressionFunction(self, xNormalized):
        """Evaluates the regression part of the model at a specific point of interest
        f(x)*beta in annas work"""
        regressVec = self._regressionVector(xNormalized)
        return regressVec @ self.beta0

    def _regressionVector(self, xNormalized):
        """Vector for evaluation of the regression model.

        For Kriging objects, the regression function can be chosen by setting the model parameter 'regress'.
        For HierarchicalKriging objects, the regression function is the Kriging model of lf Data.
        :param x: point to be evaluated
        :return: regression function value
        """
        if self.regress == "const":
            return [1]
        elif self.regress == "linear":
            regressionValues = np.ones((self.p + 1,))
            regressionValues[1:] = xNormalized
            return regressionValues
        else:
            raise DelisMMError("Only these values are applicable for parameter regress: [linear, const].")

    def _getRegressionFunctionValues(self):
        """
        Returns a matrix with the regression function values of the sample points. F = (f(x1),...,f(xn)).T
        :return: matrix with regression function values of sample points
        p. 14, top in annas work F=...
        """

        F = []
        for i in range(self.n):
            F.append(self._regressionVector(self.sampleXNormalized[:, i]))
        return np.array(F)

    def _getCorrelationMatrix(self, theta, regularizationParameter):
        """Returns the correlation matrix of the sample points

        It uses the numpy arrays to calculate each entry of R with one call. Thus it is fast but
        requires much memory. If you get a memory error, you should switch to python64 or use
        self._getCorrelationMatrixMemory which is very slow.
        :return: correlation matrix R
        """
        # ===========================================================================================
        # map sample x onto an array in axis=0 and axis=1
        # Example:
        # p=2
        # sampleX = np.array([[0.,1.,3.],[1.,2.,4.]])
        #
        # sk, sk.shape
        # [ 0.  1.] | [ 1.  2.] | [ 3.  4.]
        # [ 0.  1.] | [ 1.  2.] | [ 3.  4.]
        # [ 0.  1.] | [ 1.  2.] | [ 3.  4.]
        # (3, 3, 2)
        #
        # si, si.shape
        # [ 0.  1.] | [ 0.  1.] | [ 0.  1.]
        # [ 1.  2.] | [ 1.  2.] | [ 1.  2.]
        # [ 3.  4.] | [ 3.  4.] | [ 3.  4.]
        # (3, 3, 2)
        # ===========================================================================================
        sk = np.ones((self.n, self.n, self.p)) * self.sampleXNormalized.T
        sampleDiffs = sk.transpose((1, 0, 2)) - sk
        del sk  # remove sk to save memory

        R = self._correlationFunction(sampleDiffs, theta)
        R = R + np.identity(self.n) * regularizationParameter

        return R

    def correlationFunction(self, x, z):  # spatial correlation of two points in design space
        """Returns the correlation between x and z

        :param x: point of interest (dimension p)
        :param z: other point of interest (dimension p)
        :return: correlation value
        """
        x = np.ones((1, 1, self.p)) * x
        z = np.ones((1, 1, self.p)) * z

        return self._correlationFunction(x - z, self.theta)[0]

    def correlationFunctionNSamples(self, x, z):
        """Returns the correlation between x and z where x and z contain several samples

        :param x: points of interest (dimension (m,p))
        :param z: other points of interest (dimension (m,p))
        :return: correlation vector
        """
        x = np.array(x)
        x, z = np.ones((1, x.shape[0], self.p)) * x, np.ones((1, x.shape[0], self.p)) * z

        return self._correlationFunction(x - z, self.theta)

    def _correlationFunction(self, sampleDifferences, theta):
        """Returns the spatial correlation of each point in self.sampleXNormalized.

        You can choose the spatial correlation model by setting the model parameter 'corr'.
        :param sampleDifferences: (n,n,p)-array which is the difference as sampleXNormalized[k]-sampleXNormalized[i] where k,i= 1 .. n
        :param theta: spatial correlation parameter
        :return: spatial correlation matrix
        """
        if self.corr == "gauss":
            return np.exp(-np.sum(theta * (np.abs(sampleDifferences)) ** 2, 2))
        elif self.corr == "exp":
            return np.exp(-np.sum(theta * np.abs(sampleDifferences), 2))
        elif self.corr == "cubic":
            # using inline array replacements for memory efficiency
            sampleDifferences = theta * np.abs(sampleDifferences, sampleDifferences)
            np.fmin(sampleDifferences, np.ones_like(sampleDifferences), sampleDifferences)  # fmin used to save memory
            return np.prod(1 - 3 * sampleDifferences**2 + 2 * sampleDifferences**3, 2)
        else:
            raise DelisMMError('Surrogate parameter "corr" must be one of [gauss, exp, cubic]')

    def _getCorrelationMatrixMemory(self, theta, regularizationParameter):
        """Returns the correlation matrix of the sample points.

        It uses a python-style loop to calculate
        each entry in R. Thus it is quite slow but memory efficient.
        :return: correlation matrix R
        """
        r = np.identity(self.n) * (1 + regularizationParameter)
        for i in range(self.n):
            for k in range(i + 1, self.n):
                res = self._correlationFunctionMemory(self.sampleXNormalized[:, i], self.sampleXNormalized[:, k], theta)
                r[i, k] = res
                r[k, i] = res
        return r

    def _correlationFunctionMemory(self, x, z, theta):  # spatial correlation of two points in design space
        """Returns the spatial correlation of two points.

        You can choose the spatial correlation model by setting the model parameter 'corr'.
        :param x: first point
        :param z: second point
        :param theta: spatial correlation parameter
        :return: spatial correlation of x and z
        """
        if self.corr == "gauss":
            summand = theta * (np.abs(x - z)) ** 2
            r = np.exp(-np.sum(summand))
        elif self.corr == "exp":
            summand = theta * (np.abs(x - z))
            r = np.exp(-np.sum(summand))
        elif self.corr == "cubic":
            r = 1
            if self.p == 1:
                xij = np.amin([[1], theta * np.abs(x - z)[0]])
                xij = np.float64(xij)
                r = 1 - 3 * xij**2 + 2 * xij**3
            else:
                for j in range(self.p):
                    # the following if-clause had is needed because with p=2 design parameters there may be x-values that are floats and no array
                    #                    if type(x) is not np.ndarray:
                    #                        log.info('wrong x vector: '+str(x)+str(type(x))+'x vector set manually to length of self.p')
                    #                        x = np.zeros(self.p, np.ndarray)
                    #                    elif type(z) is not np.ndarray:log.info('wrong z vector: '+str(z)+str(type(z)))
                    #                    elif type(theta) is not np.ndarray:log.info('wrong theta vector: '+str(theta)+str(type(theta)))
                    xij = np.amin([1, theta[j] * np.abs(x[j] - z[j])])
                    rj = 1 - 3 * xij**2 + 2 * xij**3
                    r = r * rj
        else:
            raise DelisMMError("corr muss gauss, exp oder cubic sein!")
        return r

    def getCorrelationVectorMemory(self, x):
        """Returns the correlation vector of a point x and the sample points

        :param x: point of interest
        :return: correlation vector
        """
        vector = np.zeros((self.n, 1))
        for i in range(self.n):
            vector[i, 0] = self._correlationFunctionMemory(x, self.sampleXNormalized[:, i], self.theta)
        return vector

    def getCorrelationVector(self, x):
        """Returns the correlation vector of a point x and the sample points, eq. 3.15, see r(x)

        :param x: point of interest
        :return: correlation vector
        """
        si = np.ones((1, self.n, self.p)) * self.sampleXNormalized.T
        r = self._correlationFunction(si - x, self.theta)
        return r.T

    def _getModifiedCorrMatrix(self):
        """Returns the block matrix

        [ R    F
          F.T  0 ]
        where R is the correlation matrix of the sample points and F is the regression function value matrix.
        This matrix is used to evaluate the surrogate model.
        :return: matrix [[R,F];[F.T,0]]"""
        regressionValues = self._F
        if self.regress == "const":
            dim = self.n + 1
            retArray = np.zeros((dim, dim))
            retArray[self.n : dim, 0 : self.n] = regressionValues.T
            retArray[0 : self.n, self.n : dim] = regressionValues
            for i in range(self.n):
                retArray[self.n, i] = 1
                retArray[i, self.n] = 1
        elif self.regress == "linear":
            dim = self.n + self.p + 1
            retArray = np.zeros((dim, dim))
            retArray[self.n : dim, 0 : self.n] = regressionValues.T
            retArray[0 : self.n, self.n : dim] = regressionValues
        retArray[0 : self.n, 0 : self.n] = self._R
        return retArray

    def negLnLikelihoodExpOptParamWrapper(self, optParameters):
        """wrapper around the negLnLikelihood optimization target function.

        For the optimization theta is not used directly but :math:`thetaExp = log_{10}(theta)` is used instead.
        Hence, this method will wrap the thetaExp from the exponential form to the non exponential form.
        """
        optParameters = np.power(10, optParameters)
        return self.negLnLikelihood(optParameters)

    def negLnLikelihood(
        self, optParameters
    ):  # dient zur Optimierung bezueglich theta, zunaechst simplified covariance data fitment 2.5.3
        """
        Implementation of the condensed log-likelihood function, used to optimize the correlation parameter theta.

        see eq. 3.33

        :param optParameters: list of all thetas and the regularizationParameter
        :return: condensed log-likelihood function value
        """
        if self.doRegularizationParameterOpt:
            theta = optParameters[:-1]
            regularizationParameter = optParameters[-1]
        else:
            theta = optParameters
            regularizationParameter = self.regularizationParameter
        R = self._getCorrelationMatrix(theta, regularizationParameter)

        try:
            cholR = np.linalg.cholesky(R)
        except (np.linalg.linalg.LinAlgError, linalg.LinAlgWarning):
            log.debug("Matrix ist nicht positiv definit! theta = " + str(theta))
            return np.inf  # 10**20
        lnDet = np.sum(2 * np.log(cholR.diagonal()))
        try:
            solveRegress = linalg.solve(R, self._F, assume_a="pos")
        except (np.linalg.linalg.LinAlgError, linalg.LinAlgWarning):
            log.warning(
                "Regression function could not be calculated due to a singular matrix. "
                "This parameter set will be igonred. Please check your in- and output for "
                "theta = " + str(theta)
            )
            return np.inf

        solveYs = linalg.solve(R, self.sampleY, assume_a="pos")
        beta = np.linalg.solve(self._F.T @ solveRegress, self._F.T @ solveYs)
        sigmaSq = self.getSigmaSqared(R, beta, solveYs)
        result = self.n * np.log(sigmaSq) + lnDet
        return result

    @staticmethod
    def getV(F, R, Y, beta):
        """Eq 3.23 from Sauerbrei

        :param F: Regression function values eq.
        :param R: Correlation matrix
        :param Y: function values at samples
        :param symPos: flag if correlation matrix is assumed to be positive definite
        """
        v = linalg.solve(R, Y - F @ beta)
        return v

    @staticmethod
    def getBeta(F, R, Y, symPos=True):
        """Eq 3.32 from Sauerbrei

        :param F: Regression function values eq.
        :param R: Correlation matrix
        :param Y: function values at samples
        :param symPos: flag if correlation matrix is assumed to be positive definite
        """

        try:
            solveRegress = linalg.solve(R, F, assume_a="pos" if symPos else "gen")
        except np.linalg.linalg.LinAlgError:
            log.warning(
                "Regression function could not be calculated due to a singular matrix. This parameter set will be igonred."
            )
            return np.inf

        solveYs = linalg.solve(R, Y, assume_a="pos" if symPos else "gen")
        beta = np.linalg.solve(F.T @ solveRegress, F.T @ solveYs)
        return beta

    def _optimizeHyperparams(self, theta=None, doRemoteCall=False, runDir=None):
        """Performs the optimization of theta and the regularization Parameter(often called lambda, in Annas thesis it is epsilon)

        The correlation width theta is chosen for each parameter according to the number of samples, the distribution of the
        samples in the design space and the importance of the respective parameters. There is a optimization
        procedure in order to calculate theta using the target function self.negLnLikelihood.

        There are three variants of optimization runs according to parameter self.optThetaGlobalLocalType:
        - global
        - local

        Pure global or pure local optimization can be performed if the calculations take very long.
        Additionally a combined optimization can be done. In combined mode the best value of the global optimizer
        after 5 iterations is used as starting point for the local optimizer. This is repeated
        "self.optThetaGlobalAttempts"-times.
        :param theta: theta may be given as start value. If not [.2]*self.p is starting point

        :return: tuple, (optimized thetas, regularizationParameter). (Attention: it is not thetaExp!)
        """
        f = self.negLnLikelihoodExpOptParamWrapper
        bounds = self._getThetaOptBounds()
        if self.optThetaGlobalLocalType == "global":
            optResult = self._doGlobalOpt(f, bounds, doRemoteCall, runDir)
        elif self.optThetaGlobalLocalType == "local":
            optResult = self._getLocalOptimum(f, theta, bounds)
        else:
            raise DelisMMError(
                'Given value in self.optThetaGlobalLocalType is incorrect. Got "'
                + str(self.optThetaGlobalLocalType)
                + '" but just one of these values is valid: [global, local]'
            )
        log.debug(
            "Opt results (TargetValue, thetaExp):  \n"
            + indent([[self.optThetaGlobalLocalType] + [optResult.fun] + list(optResult.x)])
        )
        log.debug(
            "Opt results (TargetValue, theta):  \n"
            + indent([[self.optThetaGlobalLocalType] + [optResult.fun] + np.power(10, optResult.x).tolist()])
        )
        optResult = np.power(10, optResult.x)
        if self.doRegularizationParameterOpt:
            regularizationParameter = optResult[-1]
            theta = optResult[:-1]
        else:
            regularizationParameter = self.regularizationParameter
            theta = optResult
        return theta, regularizationParameter

    def _doGlobalOpt(self, f, bounds, doRemoteCall=False, runDir=None):
        """doc"""
        optResultHeader = (
            ["f"] + self.parameterNames + (["regularization"] if self.doRegularizationParameterOpt else [])
        )
        if self.optThetaGlobalAttempts == 1 and not doRemoteCall:
            return self._getGlobalOptimum(f, bounds)
        else:
            args = (f, bounds)
            # do parallel run
            optThetaGlobalAttempts = self.optThetaGlobalAttempts
            self.optThetaGlobalAttempts = 1
            thetaOptimizer = ThetaOptimizerKriging(self, args, doRemoteCall, runDir)
            optResults = getY([[0] * optThetaGlobalAttempts], thetaOptimizer, False)
            optValues = [result.fun for result in optResults]
            minResultIndex = np.argmin(optValues)
            optResultArray = [[optResult.fun] + list(optResult.x) for optResult in optResults]
            log.debug("multiple thetaExp opt results:\n" + indent([optResultHeader] + optResultArray))
            nonExpOptResults = np.array(optResultArray)
            nonExpOptResults[:, 1:] = np.power(10, nonExpOptResults[:, 1:])
            log.debug("multiple theta opt results:\n" + indent([optResultHeader] + nonExpOptResults.tolist()))
            self.optThetaGlobalAttempts = optThetaGlobalAttempts
            return optResults[minResultIndex]

    def _getGlobalOptimum(self, f, bounds, maxIter=None):
        """calculates the optimum of f. It performs a global and a successive local search."""
        if maxIter is None:
            maxIter = self.optThetaMaxIter
        log.info("Calculate global opt")
        return differential_evolution(f, bounds, maxiter=maxIter, disp=self.optThetaPrintout)

    def _getLocalOptimum(self, f, theta, bounds):
        """doc"""
        if theta is None:
            theta = 0.2
            thetaExp = np.log10(theta)
            startValue = [thetaExp] * self.p + (
                [self.regularizationParameter] if self.doRegularizationParameterOpt else []
            )
        else:
            thetaExp = np.log10(theta).tolist()
            startValue = thetaExp + [self.regularizationParameter]
        options = {"maxiter": 10**3, "disp": self.optThetaPrintout}
        return minimize(f, startValue, method="L-BFGS-B", bounds=bounds, options=options)

    def _getThetaOptBounds(self):
        """bounds of the correlation model depend on the type of correlation model"""
        lb = self.p * [-3.0] + ([-15.0] if self.doRegularizationParameterOpt else [])
        if self.corr == "cubic":
            ub = self.p * [0.0] + ([-1.0] if self.doRegularizationParameterOpt else [])
        else:
            ub = self.p * [2.0] + ([-1.0] if self.doRegularizationParameterOpt else [])
        return list(zip(lb, ub))

    def getParametersNormalized(self, pointOfInterest):
        if np.any(self.log10SampleX):
            pointOfInterest[self.log10SampleX] = np.log10(pointOfInterest)

        pointOfInterestNormalized = BoundsHandler.scaleTo01Static(
            np.array(pointOfInterest), self.lowerBounds, self.upperBounds
        )
        return pointOfInterestNormalized

    def __call__(self, pointOfInterest):
        """Returns the function value at pointOfInterest.

        :param pointOfInterest: vector of length self.p within real bounds"""
        return self.callNormalized(self.getParametersNormalized(pointOfInterest))

    def callNormalized(self, pointOfInterestNormalized):
        """same as __call__ but uses normalized input samples"""
        self.checkModelCreation()
        corrVec = self.getCorrelationVector(pointOfInterestNormalized)
        regressValue = self.regressionFunction(pointOfInterestNormalized)
        result = regressValue + corrVec.T @ self.v
        if self.log10SampleY:
            result = np.power(10, result)
        return result

    def getSigmaSqared(self, R=None, beta=None, solveYs=None):
        """Calculates sigmasquared from eq 3.30"""
        if R is None:
            R = self._R
        if beta is None:
            beta = self.beta0
        if solveYs is None:
            solveYs = linalg.solve(R, self.sampleY, assume_a="pos")

        solveRegressValues = linalg.solve(R, self._F, assume_a="pos")
        vector = self.sampleY - self._F @ beta
        sigmaSq = (vector.transpose() @ (solveYs - solveRegressValues @ beta)) / self.n
        return sigmaSq

    def getMSEofPredictor(self, pointOfInterestNormalized):
        """Returns the predicted MSE of the Kriging model

        :param pointOfInterestNormalized: point to be evaluated
        :return: predicted MSE at pointOfInterestNormalized"""
        sigmaSq = self.getSigmaSqared()
        solveRegressValues = np.linalg.solve(self._R, self._F)
        corrVec = self.getCorrelationVector(pointOfInterestNormalized)
        solveCorrVec = np.linalg.solve(self._R, corrVec)
        regresssionVector = self._regressionVector(pointOfInterestNormalized)
        vector = corrVec.T @ solveRegressValues - regresssionVector
        mse = sigmaSq * (
            1.0 - corrVec.T @ solveCorrVec + vector * np.linalg.solve(self._F.T @ solveRegressValues, vector.T)
        )
        return mse[0, 0]

    def getExpectedImprovement(self, pointOfInterestNormalized):
        """Calculates the expected improvement function

        Ref: see Forrester et al. [FSK08, Kap. 3] or Diss Freund eq. 3.113

        :param pointOfInterestNormalized: point to be evaluated
        :return: predicted EIF at pointOfInterestNormalized
        """
        s = np.sqrt(self.getMSEofPredictor(pointOfInterestNormalized))
        yMin = np.min(self.sampleY)
        yHat = self.callNormalized(pointOfInterestNormalized)
        yDiff = (yMin - yHat)[0]
        a = yDiff / s

        left = 0.5 * (yDiff) * (1 + special.erf(a / np.sqrt(2)))
        right = s / np.sqrt(2 * np.pi) * np.exp(-0.5 * a**2)

        return left + right

    def checkModelCreation(self):
        """Raises an error, if the model is not created. Returns otherwise"""
        if not self.isCreated:
            raise DelisMMError(
                "The kriging model is not created! Please run createSurrogateModel() prior to this call."
            )

    def getCopyNewSamples(self, sampleXNormalized, sampleY):
        """Returns a copy of self but with adjusted samples. Thus n and p are adjusted as well"""
        kriging = copy.copy(self)
        kriging.sampleXNormalized = sampleXNormalized
        kriging.sampleY = np.array(sampleY)
        kriging.n = sampleXNormalized.shape[1]  # n - Number of sample points
        kriging.p = sampleXNormalized.shape[0]  # p - Number of design variables
        return kriging

    def getKrigingDocumentationDict(self):
        """Returns a OrderedDict with"""
        attributesToDocument = [
            "regularizationParameter",
            "optThetaGlobalLocalType",
            "optThetaGlobalAttempts",
            "optThetaMaxIter",
            "optThetaPrintout",
            "correlationMatrixAlgo",
            "log10SampleX",
            "log10SampleY",
        ]
        docDict = OrderedDict(
            [
                ("kriging_type", self.__class__.__name__),
                ("model_name", self.name),
                ("number_of_samples", self.n),
                ("number_of_parameters", self.p),
                ("condition", self.condition),
                ("correlation_function", self.corr),
                ("regression_function", self.regress),
            ]
            + [(attributeName + "", getattr(self, attributeName)) for attributeName in attributesToDocument]
        )

        if isinstance(self, HierarchicalKriging):
            lowFiModelName = (
                self.lfKriging.name if hasattr(self.lfKriging, "name") else self.lfKriging.__class__.__name__
            )
            docDict.update({"low fi model", lowFiModelName})

        listsToAdd = ["parameterNames", "lowerBounds", "upperBounds", "theta"]
        docDict.update([(listIt, list(getattr(self, listIt))) for listIt in listsToAdd])
        return SurrogateInfo(**docDict)

    def getKrigingDocumentation(self):
        """doc"""
        docDict = self.getKrigingDocumentationDict().model_dump()
        parameterNames = docDict.pop("parameterNames")
        lowerBounds = docDict.pop("lowerBounds")
        upperBounds = docDict.pop("upperBounds")
        theta = docDict.pop("theta")
        keys = [key + ":" for key in docDict.keys()]
        documentation = list(zip(keys, docDict.values()))

        theta = [None] * self.p if theta is None else theta
        boundsArray = [["parameterNames", "lowerBounds", "upperBounds", "thetas"]] + list(
            zip(parameterNames, lowerBounds, upperBounds, theta)
        )

        allSamples = []
        sampleY = np.power(10, self.sampleY) if self.log10SampleY else self.sampleY
        for inputSample, outputSample in zip(self.sampleX.T, sampleY):
            if hasattr(outputSample, "__iter__"):
                allSamples.append(list(inputSample) + list(outputSample))
            else:
                allSamples.append(list(inputSample) + list([outputSample]))

        delim = " "
        return (
            indent(documentation, delim=delim)
            + "\n\nBounds:\n"
            + indent(boundsArray, delim=delim)
            + "\n\nSamples:\n"
            + indent([self.parameterNames + self.resultNames] + allSamples, delim=delim)
        )

    def isCorrectVersion(self):
        """doc"""
        if isinstance(self.version, float) and abs(self.version - 1.1) < 1e-8:
            log.info("Read kriging model of old version. Transform kriging model v1.1 to v1.2")
            self.regularizationParameter = 1e-11  # in v1.1 the regularizationParameter was a constant set to 1e-11
            self.version = "1.2"
        if self.version == "1.2":
            log.info("Read kriging model of old version. Transform kriging model v1.2 to v1.3")
            self.doRegularizationParameterOpt = True
            self.version = "1.3"
        if self.version == "1.3":
            log.info("Read kriging model of old version. Transform kriging model v1.3 to v1.4")
            self.version = "1.4"  # only the inputs and outputs changed
        if self.version == "1.4":
            log.info("Read kriging model of old version. Transform kriging model v1.4 to v1.5")
            oldMatrixAttributes = ["_F", "_R", "_modifiedCorrMatrix", "beta0"]
            for att in oldMatrixAttributes:
                setattr(self, att, np.array(getattr(self, att)))
            self.beta0 = self.getBeta(self._F, self._R, self.sampleY)
            self.v = self.getV(self._F, self._R, self.sampleY, self.beta0)
            self._ys = None  # not used anymore
            self.version = "1.5.0"  # only the inputs and outputs changed
        if self.version == "1.5.0":
            log.info("Read kriging model of old version. Transform kriging model v1.5.0 to v1.6.0")
            self.log10SampleY = False
            self.log10SampleX = np.array([False] * len(self.parameterNames))
            self.version = "1.6.0"

        if self.version.rsplit(".", 1)[0] != self.__class__.version.rsplit(".", 1)[0]:
            log.error(
                "The version of the surrogates read does not correspond to the actual surrogate implementation. "
                + "Actual program version, loaded surrogate version: "
                + str(self.__class__.version)
                + ", "
                + str(self.version)
            )
            return False
        return True

    def save(self, filename, path=""):
        """saves the surrogate to the file in filename."""
        # remove large arrays. They can be recreated easily with self.theta
        temp = self._F, self._R, self._modifiedCorrMatrix, self.v
        self._F, self._R, self._modifiedCorrMatrix, self.v = [None] * 4
        super(AbstractKriging, self).save(filename, path)
        self._F, self._R, self._modifiedCorrMatrix, self.v = temp

    def _isCreated(self):
        """Checks if self.createSurrogateModel was run and the required values are set"""
        return self.v is not None and self.theta is not None

    isCreated = property(fget=_isCreated)


class Kriging(AbstractKriging):
    """Kriging class - takes at least two arguments: sampleX, sampleY.

    :param sampleX: is an array in which contains the sample points column by column (one parameter configuration = one column)
        Attention: sampleX is scaled to bounds. This module normalizes sampleX itself
    :param sampleY: is an array in which contains the scalar function values at the sample points
    :param lowerBounds: array with lower bounds of length self.p. Defaults to [0, 0, ...]
    :param upperBounds: array with upper bounds of length self.p. Defaults to [1, 1, ...]

    Optionally you can determine by the 3rd argument corr which correlation model will be used. By default, corr is 'cubic', for cubic correlation model,
    which is the numerically most stable one. But you can also choose 'exp' for exponential correlation model or 'gauss' for gaussian correlation model.
    By setting the 4th argument regress, you can determine which regression model will be used. It is 'const' by default, but you can also choose 'linear'
    for a linear regression.
    """

    def __init__(
        self,
        sampleX,
        sampleY,
        lowerBounds=None,
        upperBounds=None,
        corr="cubic",
        regress="const",
        parameterNames=None,
        resultNames=None,
        **kwargs,
    ):
        AbstractKriging.__init__(
            self, sampleX, sampleY, lowerBounds, upperBounds, corr, regress, parameterNames, resultNames
        )

        for key in kwargs:
            if not hasattr(self, key):
                raise DelisMMError(
                    'Setting unknown key "'
                    + key
                    + '" in class "'
                    + str(self.__class__)
                    + '" with name "'
                    + str(self)
                    + '"'
                )
            setattr(self, key, kwargs[key])


class HierarchicalKriging(AbstractKriging):
    """
    Hierarchical Kriging class - takes at least three arguments: sampleX, sampleY, lfKriging.
    :param sampleX: is an array in which contains the sample points column by column (one parameter configuration = one column)
        Attention: sampleX is scaled to bounds. This module normalizes sampleX itself
    :param sampleY: is an array in which contains the scalar function values at the sample points
    :param lfKriging: is an instance of Kriging or HierarchicalKriging class; model for the low-fidelity data.
    :param lowerBounds: array with lower bounds of length self.p. Defaults to [0, 0, ...]
    :param upperBounds: array with upper bounds of length self.p. Defaults to [1, 1, ...]

    Optionally you can determine by the 3rd argument corr which correlation model will be used.
    By default, corr is 'cubic', for cubic correlation model, which is the numerically most stable one.
    But you can also choose 'exp' for exponential correlation model or 'gauss' for gaussian correlation model.
    By setting the 4th argument regress, you can determine which regression model will be used.
    It is 'const' by default, but you can also choose 'linear'
    for a linear regression.
    """

    def __init__(
        self,
        sampleX,
        sampleY,
        lfKriging,
        lowerBounds=None,
        upperBounds=None,
        corr="cubic",
        regress="const",
        parameterNames=None,
        resultNames=None,
        **kwargs,
    ):
        self.lfKriging = lfKriging
        AbstractKriging.__init__(
            self,
            sampleX,
            sampleY,
            lowerBounds,
            upperBounds,
            corr,
            regress,
            parameterNames=parameterNames,
            resultNames=resultNames,
        )
        self.name = "hier-" + self.name

        for key in kwargs:
            if not hasattr(self, key):
                raise DelisMMError(
                    'Setting unknown key "'
                    + key
                    + '" in class "'
                    + str(self.__class__)
                    + '" with name "'
                    + str(self)
                    + '"'
                )
            setattr(self, key, kwargs[key])

    def _regressionVector(self, xNormalized):
        """See AbstractKriging for documentation"""
        regressionVector = super()._regressionVector(xNormalized)
        regressionVector[0] = self.lfKriging.callNormalized(xNormalized)[0]
        return regressionVector


if __name__ == "__main__":
    pass



"""
Created on 24.03.2015

@author: freu_se


This module contains error metrics
"""

import numpy as np

from delismm.service.exception import DelisMMError


def checkSamples(sample1, sample2, removeNHighest=0):
    """doc"""

    if not isinstance(sample1, np.ndarray):
        sample1 = np.array(sample1)
    if not isinstance(sample2, np.ndarray):
        sample2 = np.array(sample2)
    if removeNHighest > 0 and removeNHighest >= len(sample1):
        raise DelisMMError("Could not remove n-highest input sample. The sample array is equal or smaller than n.")
    if not isinstance(removeNHighest, int):
        raise DelisMMError("removeNHighest must be of type int")
    if len(sample1) != len(sample2):
        raise DelisMMError("Input samples do not have equal size!")
    return sample1, sample2


def meanSquareError(sample1, sample2, removeNHighest=0):
    """
    .. math::
        MSE=\\frac{1}{n}\\sum_{i=1}^n\\left(\\hat{Y_i} - Y_i\\right)^2
    """
    sample1, sample2 = checkSamples(sample1, sample2, removeNHighest)
    result = (sample1 - sample2) ** 2
    if removeNHighest > 0:
        result = np.sort(result)[: -1 * removeNHighest]
    return result.mean()


def rootMeanSquareError(sample1, sample2, removeNHighest=0):
    """
    .. math::
        RMSE=\\sqrt{\\frac{1}{n}\\sum_{i=1}^n\\left(\\hat{Y_i} - Y_i\\right)^2}
    """
    sample1, sample2 = checkSamples(sample1, sample2, removeNHighest)
    return np.sqrt(meanSquareError(sample1, sample2, removeNHighest))


def relativeMeanSquareError(sample1, sample2, removeNHighest=0):
    """
    .. math::
        MSRE=\\frac{1}{n}\\sum_{i=1}^n\\left(1 - \\frac{min(\\hat{Y_i}, Y_i)}{max(\\hat{Y_i},Y_i)}\\right)^2
    """
    sample1, sample2 = np.abs(checkSamples(sample1, sample2, removeNHighest))
    minByMax = np.array(np.min(np.vstack((sample1, sample2)), 0), dtype=np.float64) / np.max(
        np.vstack((sample1, sample2)), 0
    )
    return meanSquareError(np.ones_like(minByMax), minByMax, removeNHighest)


def relativeRootMeanSquareError(sample1, sample2, removeNHighest=0):
    """
    .. math::
        RMSRE=\\sqrt{\\frac{1}{n}\\sum_{i=1}^n\\left(1 - \\frac{min(\\hat{Y_i}, Y_i)}{max(\\hat{Y_i},Y_i)}\\right)^2}
    """
    sample1, sample2 = np.abs(checkSamples(sample1, sample2, removeNHighest))
    minByMax = np.min(np.array([sample1, sample2]), axis=0) / np.max(np.array([sample1, sample2]), axis=0)
    return rootMeanSquareError(np.ones_like(minByMax), minByMax, removeNHighest)


def normalizedRootMeanSquareError(sample1, sample2):
    """
    .. math::
        NRMSE=\\frac{\\sqrt{\\frac{1}{n}\\sum_{i=1}^n(\\hat{Y_i}-Y_i)^{2}}}{max(abs(Y))}

    :param sample1: sample points used to create the metamodel (Y_i below)
    :param sample2: cross validation values (\\hat{Y_i} below)
    :return: normalized root mean square error
    """
    sample1, sample2 = np.abs(checkSamples(sample1, sample2, removeNHighest=0))
    lenY = len(sample1)
    mse = sum([(sample1[i] - sample2[i]) ** 2 for i in range(lenY)]) / lenY
    max_y = max([abs(sample1[i]) for i in range(lenY)])
    nrmse = np.sqrt(mse) / max_y
    return nrmse


def meanRelativeError(sample1, sample2, removeNHighest=0):
    """
    .. math::
        MRE=\\frac{1}{n}\\sum_{i=1}^n 1 - \\frac{min(\\hat{Y_i}, Y_i)}{max(\\hat{Y_i},Y_i)}

    :param sample1: sample points used to create the metamodel (Y_i below)
    :param sample2: cross validation values (\\hat{Y_i} below)
    :return: mean relative error
    """
    sample1, sample2 = np.abs(checkSamples(sample1, sample2, removeNHighest=0))
    minByMax = np.array(np.min(np.vstack((sample1, sample2)), 0), dtype=np.float64) / np.max(
        np.vstack((sample1, sample2)), 0
    )
    mre = np.mean(minByMax)
    return np.abs(1 - mre)


def r2pressCov(sample1, sample2, removeNHighest=0):
    """Calculates the covariance based r2pres value

    :param sample1: sample points used to create the metamodel (Y_i below)
    :param sample2: cross validation values (\\hat{Y_i} below)

    For a description see
    Forrester p 37

    .. math::
        R^2_{press}=1-(\\frac{cov(s1,s2)}{\\sqrt{var(s1)var(s2)}})^2
    """
    sample1, sample2 = checkSamples(sample1, sample2, removeNHighest)
    cov = np.mean(sample1 * sample2) - np.mean(sample1) * np.mean(sample2)
    r2Corr = (cov / np.sqrt(np.var(sample1) * np.var(sample2))) ** 2
    return r2Corr


def r2press(sample1, sample2, removeNHighest=0):
    """Calculates the r2pres value from Optimus

    :param sample1: sample points used to create the metamodel (Y_i below)
    :param sample2: cross validation values (\\hat{Y_i} below)

    For a description see
    Optimus Theroy Guide Rev10_14 chapter 1.3.3 "Error measures and regression parameters"

    Basically it uses a cross correlation of all sample points used. These are used in the
    SSE. Additionally the SSTOT is calculated to account for the distribution of the samples.

    sum of squared errors:
    .. math::
        SSE=\\sum_{i=1}^n\\left(\\hat{Y_i} - Y_i\\right)^2

    total sum of sqares(sum of squared deviation from the mean value):
    .. math::
        SSTOT=\\sum_{i=1}^n\\left(Y_i - \\bar{Y_i}\\right)^2


    .. math::
        R^2_{press}=1-\\frac{SSE}{SSTOT}
    """
    return 1 - r2pressSub1(sample1, sample2, removeNHighest)


def r2pressSub1(sample1, sample2, removeNHighest=0):
    """substract r2pres from 1

    .. math::
        1 - R^2_{press}=\\frac{SSE}{SSTOT}

    See r2pres for parameter description"""
    sample1, sample2 = checkSamples(sample1, sample2, removeNHighest)
    if removeNHighest > 0:
        # sample1 needs to be reduced as well. Thus it is included in the sorting array
        sampleArray = [(diffs, s1) for diffs, s1 in zip(np.abs(sample1 - sample2), sample1)]
        sampleArray.sort(key=lambda x: x[0])
        sampleArray = np.array(sampleArray[: -1 * removeNHighest]).T
        result, sample1 = sampleArray
    else:
        result = np.abs(sample1 - sample2)
    result = (result**2).mean()
    sse = len(sample1) * result
    sstot = np.sum(np.square(sample1 - np.mean(sample1)))
    r2 = sse / sstot
    return r2


def maxScaledError(sample1, sample2, removeNHighest=None):
    """Calculate the maximum error from cross validation results scaled to the interval size

    :param sample1: sample points used to create the metamodel (Y_i below)
    :param sample2: cross validation values (\\hat{Y_i} below)
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = checkSamples(sample1, sample2)
    diffMinMax = np.max(sample1) - np.min(sample1)
    maxCrossValidationDifference = np.array(np.max(np.abs(sample1 - sample2)), np.float64)
    return maxCrossValidationDifference / diffMinMax


def maxScaledErrorLog10(sample1, sample2, removeNHighest=None):
    """calculate log_10(sample1) and log_10(sample2) and then use maxScaledError

    :param sample1: sample points used to create the metamodel (Y_i below)
    :param sample2: cross validation values (\\hat{Y_i} below)
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = checkSamples(sample1, sample2)
    return maxScaledError(np.log10(sample1), np.log10(sample2))


def maxRelativeError(sample1, sample2, removeNHighest=None):
    """Calculate the maximum relative error from cross validation results

    .. math::
        MaxRE \\coloneqq max\\left(1 - \\frac{min\\left(\\hat{Y_i}, Y_i\\right)}{max\\left(\\hat{Y_i},Y_i\\right)}\\right)

    :param sample1: sample points used to create the metamodel
    :param sample2: cross validation values
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = np.abs(checkSamples(sample1, sample2))
    minByMax = np.array(np.min(np.vstack((sample1, sample2)), 0), dtype=np.float64) / np.max(
        np.vstack((sample1, sample2)), 0
    )
    return np.max(1 - minByMax)


def maxRelativeErrorLog10(sample1, sample2, removeNHighest=None):
    """calculate log_10(sample1) and log_10(sample2) and then use maxRelativeError

    :param sample1: sample points used to create the metamodel
    :param sample2: cross validation values
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = checkSamples(sample1, sample2)
    return maxRelativeError(np.log10(sample1), np.log10(sample2))


def maxError(sample1, sample2, removeNHighest=None):
    """Calculate the maximum error from two sets of sample points

    .. math::
        MaxE \\coloneqq max\\left(abs\\left(\\hat{Y_i} - Y_i\\right)\\right)

    :param sample1: sample points used to create the metamodel
    :param sample2: cross validation values
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = np.abs(checkSamples(sample1, sample2))
    return np.max(np.abs(sample1 - sample2))


def maxErrorLog10(sample1, sample2, removeNHighest=None):
    """calculate log_10(sample1) and log_10(sample2) and then use maxError

    :param sample1: sample points used to create the metamodel
    :param sample2: cross validation values
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = checkSamples(sample1, sample2)
    return maxError(np.log10(sample1), np.log10(sample2))


def correlation(sample1, sample2, removeNHighest=0):
    """calculate log_10(sample1) and log_10(sample2) and then use maxError

    :param sample1: sample points used to create the metamodel
    :param sample2: cross validation values
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = checkSamples(sample1, sample2)
    if removeNHighest > 0:
        sample1, sample2 = zip(*[(s1, s2) for _, s1, s2 in sorted(zip(np.abs(sample1 - sample2), sample1, sample2))])
        sample1, sample2 = sample1[: -1 * removeNHighest], sample2[: -1 * removeNHighest]
    return np.corrcoef(sample1, sample2)[0, 1]


def correlationLog10(sample1, sample2, removeNHighest=0):
    """calculate log_10(sample1) and log_10(sample2) and then use maxError

    :param sample1: sample points used to create the metamodel
    :param sample2: cross validation values
    :param removeNHighest: has no effect, just for coherence with all other methods here
    """
    sample1, sample2 = checkSamples(sample1, sample2)
    return correlation(np.log10(sample1), np.log10(sample2))


def bayesianInformationCriterion(sample1, sample2, k):
    """calculate bayesian information criterion for gaussian modeled error disturbances
    https://en.wikipedia.org/wiki/Bayesian_information_criterion#Gaussian_special_case

    :param sample1: sample points used to create the metamodel
    :param sample2: cross validation values
    :param k: number of paramerters that comprise the model
    """
    sample1, sample2 = checkSamples(sample1, sample2)
    n = len(sample1)
    return n * np.log(np.var(sample1 - sample2)) + k * np.log(n)


if __name__ == "__main__":
    filename = "D:\\foo.txt"

    arr = np.loadtxt(filename)

    sample = arr[:, 0]
    hfmm = arr[:, 2]
    vfmm = arr[:, 3]
    n = len(sample)

    b = bayesianInformationCriterion
    print(np.var(sample - hfmm))
    x = sample - hfmm
    print(meanSquareError(x, np.ones_like(x) * x.mean()))
    x = sample - hfmm
    print(np.mean(np.abs(x - x.mean()) ** 2))

    print(b(sample, hfmm, 8 * 2), np.var((sample - hfmm)))
    print(b(sample, vfmm, 8 * 4), np.var((sample - vfmm)))



# Copyright (C) 2013 Deutsches Zentrum fuer Luft- und Raumfahrt(DLR, German Aerospace Center) <www.dlr.de>
"""
Delismm is a pure python package for the creation of kriging-based metamodels.

It covers the following features:

- Perform parameter sensitivity analysis
- Generate Design of Experiments (DOE)
- Save/Load DOE
- Create Kriging Metamodels
- Create hierarchical kriging metamodels
- Perform a resampling to generate new Designs and improve the model
- Run DOEs using various parallelization methods local and remote


A simple example:

   >>> from delismm.example import runExample
   >>> runExample(doPlot = False)

This will create a

- doe
- sample values
- kriging model
- a diagram (if doPlot is active)


"""

from pathlib import Path

import importlib_metadata
from patme import getPyprojectMeta

name = Path(__file__).parent.name


description = ""

if Path(__file__).parts[-4] == name:
    # We have the full GitLab repository
    pkgMeta = getPyprojectMeta(__file__)
    version = str(pkgMeta["version"])
    description = str(pkgMeta["description"])
    programDir = str(Path(__file__).parents[2])
else:
    # package is installed
    version = importlib_metadata.version(name)
    programDir = str(Path(__file__).parent)
    description = importlib_metadata.metadata(name)["Summary"]
